r"""
Joint Moments of CPPs :math:`I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^{*}`

In this module, we present the computation of the joint moments of
(:math:`I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^*`). This is accomplished by deriving the
joint moment-generating function (MGF) of these processes.

The joint MGF of (:math:`I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^*`) can be derived as follows:

.. math::

    \begin{align*}
        M_{I\!E\!Z_{s,t},I\!Z_{s,t},I\!Z_{s,t}^*}(\boldsymbol{a})
        &\mathrel{:=} \mathbb{E}[e^{a_1I\!E\!Z_{s,t} + a_2I\!Z_{s,t} + a_3I\!Z_{s,t}^*}]\\
        %&= \mathbb{E}[\mathbb{E}[e^{a_1 \sum_{i=1}^{N(t-s)}e^{ks_i}J_i + a_2\sum_{i=1}^{N(t-s)}J_i + a_3\sum_{i=1}^{N(t-s)}J_i^*}|N(t-s) = n]]\\
        &= \mathbb{E}[\mathbb{E}[e^{\sum_{i=1}^n [(a_1 e^{ks_i} + a_2)J_i + a_3J_i^*]} |N(t-s) = n]]\\
        &= \mathbb{E}[\mathbb{E}[\left(\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]\cdot M_{J_i^*}(a_3)\right)^n |N(t-s) = n]]\\
        &= \sum_{n=0}^{\infty} \left(\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)] \cdot M_{J_i^*}(a_3)\right)^n \frac{[\lambda (t-s)]^n e^{-\lambda (t-s)}}{n!}\\
        &= e^{-\lambda (t-s)}  \sum_{n=0}^{\infty}  \left[\lambda (t-s)\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]\cdot M_{J_i^*}(a_3)\right]^n/n!\\
        %&= e^{-\lambda (t-s)} e^{\lambda (t-s)\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]\cdot M_{J_i^*}(a_3)}\\
        &= e^{\lambda (t-s)\left(\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]\cdot M_{J_i^*}(a_3) - 1\right)},
    \end{align*}

where :math:`\boldsymbol{a}` is a vector of real numbers, i.e.,
:math:`\boldsymbol{a} \mathrel{:=} (a_1, a_2, a_3)`, :math:`N(\cdot)` is the shared
counting process for (:math:`I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^*`), and
:math:`M_{J_i}(\cdot)` and :math:`M_{J_i^*}(\cdot)`
are the MGFs of :math:`J_i^*` and :math:`J_i`, respectively. Note that, conditional
on :math:`N(t-s) = n`, the unsorted arrival times are uniformly distributed over the
interval :math:`(s,t]`. For notational simplicity, we use :math:`\{s_1,\dots, s_n\}`
to denote these unsorted arrival times and omit the conditional notation :math:`|N(t-s) = n`
in the expectation. In what follows, we will demonstrate that
:math:`\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]` admits a closed-form expression.

First, we substitute the MGF of :math:`J_i` (conditioned on :math:`s_i`) with its known
formula:

.. math::

    \begin{align*}
        \mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]
        = \mathbb{E}[\mathbb{E}[e^{(a_1e^{ks_i} + a_2)J_i}|s_i]]
        = \mathbb{E}\left[\frac{1}{1-(a_1e^{ks_i} + a_2 )\mu_v} \right].
    \end{align*}

By introducing two new variables :math:`a \mathrel{:=} - a_1\mu_v, b \mathrel{:=} 1 - a_2\mu_v`,
the above expectation simplifies to:

.. math::

    \begin{align*}
        \mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]
        = \mathbb{E}\left[\frac{1}{ae^{ks_i} + b}\right]
        = \frac{1}{t-s}\int_{s}^t\frac{1}{ae^{ks_i}+b}\mathrm{d} s_i.
    \end{align*}

The integral can be computed explicitly using the variable substitution method. Let us
introduce :math:`x \mathrel{:=} e^{ks_i}`. Then, :math:`s_i = \frac{1}{k}\log x`,
:math:`\mathrm{d} s_i = \frac{1}{k}\frac{1}{x}\mathrm{d} x`.
The integral thus becomes:

.. math::

    \begin{align*}
        \int_{s}^t\frac{1}{ae^{ks_i}+b}\mathrm{d} s_i
        = \int_{e^{ks}}^{e^{kt}} \frac{1}{ax + b} \frac{1}{k} \frac{1}{x} \mathrm{d} x
        = \frac{1}{k}\int_{e^{ks}}^{e^{kt}} \frac{1}{ax + b} \frac{1}{x} \mathrm{d} x.
    \end{align*}

If :math:`a = 0`, i.e., :math:`a_1 = 0`, the integral simplifies to:

.. math::

    \begin{equation*}
        \int_{s}^t\frac{1}{ae^{ks_i}+b}\mathrm{d} s_i = \frac{1}{b}(t-s).
    \end{equation*}

Otherwise, for :math:`a_1\neq 0`, in any neighborhood of the origin of the vector
:math:`(a_1,a_2)` such that  :math:`ax + b \approx 1`, we have:

.. math::

    \begin{align*}
        \int_{e^{ks}}^{e^{kt}} \frac{1}{ax + b} \frac{1}{x} \mathrm{d} x
        %&= \int_{e^{ks}}^{e^{kt}} \left(\frac{-a/b}{ax + b} + \frac{1/b}{x}\right) dx\\
        %&= -\frac{1}{b}\int_{e^{ks}}^{e^{kt}} \frac{1}{ax+b}\mathrm{d}(ax+b) + \frac{1}{b}\int_{e^{ks}}^{e^{kt}}\frac{1}{x}\mathrm{d} x\\
        &= -\frac{1}{b}\left[\log(ae^{kt} + b) - \log(ae^{ks} + b)\right] + \frac{1}{b}k(t-s).
    \end{align*}

Therefore, the integral evaluates to:

.. math::

    \begin{align*}
        \int_{s}^t\frac{1}{ae^{ks_i}+b}\mathrm{d} s_i
        &= -\frac{1}{kb}\left[\log(ae^{kt} + b) - \log(ae^{ks} + b)\right] + \frac{1}{b}(t-s),
    \end{align*}

provided that :math:`(a_1,a_2)` lies within a sufficiently small neighborhood of
the origin :math:`(0,0)`. Finally, we obtain the closed-form expression for
:math:`\mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]` as:

.. math::

    \begin{equation}
        \mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]
        = \frac{1}{b}\left[1 -\frac{1}{k(t-s)}\left(\log(ae^{kt} + b) - \log(ae^{ks} + b)\right)\right],
    \end{equation}

where :math:`a = -a_1\mu_v`, :math:`b = 1 - a_2\mu_v`, and :math:`(a_1,a_2)` is restricted
to a small neighborhood around the origin :math:`(0,0)`.

To simplify the notation, we define a new function :math:`M_{E\!J,J,J^*}(\cdot)` as the
following product:

.. math::

    \begin{equation*}
        M_{E\!J,J,J^*}(\boldsymbol{a}) \mathrel{:=} \mathbb{E}[M_{J_i}(a_1e^{ks_i} + a_2)]\cdot M_{J_i^*}(a_3).
    \end{equation*}

We know that the MGF of :math:`J_i^*` has the following expression:

.. math::

    \begin{equation}
        M_{J_i^*}(a_3) = e^{\mu_s a_3 + \sigma_s^2a_3^2/2}
    \end{equation}

since :math:`J_I^*` is normally distributed with mean :math:`\mu_s` and variance
:math:`\sigma_s^2`. Combining these results, we obtain the following closed-form
expression for the joint MGF of :math:`(I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^*)`:

.. math::

    \begin{equation}
        M_{I\!E\!Z_{s,t},I\!Z_{s,t},I\!Z_{s,t}^*}(\boldsymbol{a})
        = e^{\lambda (t-s) (M_{E\!J,J,J^*}(\boldsymbol{a})-1)},
    \end{equation}

where :math:`\boldsymbol{a} = (a_1,a_2,a_3)`, :math:`a = -a_1\mu_v`, :math:`b = 1-a_2\mu_v`,
and

.. math::

    \begin{equation*}
        M_{E\!J,J,J^*}(\boldsymbol{a})
        = \frac{1}{b} \left[ 1 - \frac{1}{k(t-s)}\left(\log(ae^{kt} + b) - \log(ae^{ks} + b)\right) \right] e^{\mu_s a_3 + \sigma_s^2 a_3^2/2}.
    \end{equation*}

Given the joint MGF of :math:`(I\!E\!Z_{s,t}, I\!Z_{s,t}, I\!Z_{s,t}^*)`, we can compute
the joint moment of :math:`(I\!E\!Z_{s,t}, I\!Z_{s,t},I\!Z_{s,t}^*)` of any order.

The :math:`n`-th (:math:`n\ge1`) partial derivative of :math:`M_{E\!J,J,J^*}(\boldsymbol{a})`
with respect to :math:`a_1` is given by

.. math::

    \begin{equation*}
        \frac{\partial^nM_{E\!J,J,J^*}}{\partial a_1^{n}}
        =  \frac{1}{b}\left[\frac{e^{nkt}}{(ae^{kt}+b)^{n}}  - \frac{e^{nks}}{(ae^{ks}+b)^{n}} \right] \frac{(n-1)! \mu_v^n}{k(t-s)} M_{J_i^*}(a_3).
    \end{equation*}

Consequently, the :math:`n`-th moment of :math:`E\!J` can be expressed as

.. math::

    \begin{equation}
        \mathbb{E}[(e^{ks_i}J_i)^n] = (e^{nkt} - e^{nks})\frac{1}{k(t-s)}(n-1)!\mu_v^n.
    \end{equation}


For :math:`n_1\ge 1` and :math:`n_2\ge 1`, the following formula holds:

.. math::

    \begin{align}
        \frac{\partial^{n_1+n_2}M_{E\!J,J,J^*}}{\partial a_1^{n_1}\partial a_2^{n_2}}
        &= \sum_{i=0}^{n_2} \frac{c(n_1,n_2,i)}{b^{n_2-i+1}}\left[\frac{e^{n_1kt}}{(ae^{kt}+b)^{n_1+i}} - \frac{e^{n_1ks}}{(ae^{ks}+b)^{n_1+i}} \right] \frac{\mu_v^{n_1+n_2}}{k(t-s)} M_{J_i^*}(a_3),
    \end{align}

where :math:`c(n_1,n_2,i) \mathrel{:=} n_2!(n_1 - 1 + i)!/(i!)`. An alternative
approach involves directly computing the joint moment directly, yielding:

.. math::

    \begin{equation}
        \mathbb{E}[(e^{ks_i}J_i)^{n_1}J_i^{n_2}]
        = \mathbb{E}[e^{n_1ks_i}]\mathbb{E}[J_i^{n_1+n_2}]
        = \frac{1}{n_1k(t-s)}(e^{n_1kt} - e^{n_1ks})(n_1+n_2)!\mu_v^{n_1+n_2}.
    \end{equation}

For :math:`n\ge 1`, we have:

.. math::

    \begin{align*}
        \frac{\partial^{n}M_{E\!J,J,J^*}}{\partial a_2^{n}}
        &= \frac{n!}{b^{n+1}}  \left[ 1 - \frac{1}{k(t-s)}\left(\log(ae^{kt} + b) - \log(ae^{ks} + b)\right) \right]\mu_v^n M_{J_i^*}(a_3)\\
        &\quad + \sum_{i=1}^n\binom{n}{i}\frac{(n-i)!(i-1)!}{b^{n-i+1}} \left[\frac{1}{(ae^{kt}+b)^i} - \frac{1}{(ae^{ks} + b)^i}\right]\frac{1}{k(t-s)} \mu_v^n M_{J_I^*}(a_3)\\
        &= \frac{n!}{b^{n+1}}  \left[ 1 - \frac{1}{k(t-s)}\left(\log(ae^{kt} + b) - \log(ae^{ks} + b)\right) \right]\mu_v^n M_{J_i^*}(a_3)\\
        &\quad + \sum_{i=1}^n\frac{n!/i}{b^{n-i+1}} \left[\frac{1}{(ae^{kt}+b)^i} - \frac{1}{(ae^{ks} + b)^i}\right]\frac{1}{k(t-s)} \mu_v^n M_{J_I^*}(a_3).
    \end{align*}

Thus, the :math:`n`-th moment of :math:`J` is given by

.. math::

    \begin{equation}
        \mathbb{E}[J_i^n] = n!\mu_v^n.
    \end{equation}

"""
import math
from fractions import Fraction

from ajdmom import Poly
from ajdmom.cpp_mom import mnorm


def moment_ejjj(n1, n2, n3):
    if n1 < 0 or n2 < 0 or n3 < 0:
        raise ValueError('n1 and n2 and n3 must be non-negative integers!')
    poly = Poly()
    kf = ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v']
    poly.set_keyfor(kf)
    if n1 >= 1 and n2 == 0:
        key = (n1, 0, -1, 1, 0, n1)
        val = Fraction(math.factorial(n1 - 1), 1)
        poly.add_keyval(key, val)
        key = (0, n1, -1, 1, 0, n1)
        poly.add_keyval(key, -val)
    if n1 >= 1 and n2 >= 1:
        key = (n1, 0, -1, 1, 0, n1 + n2)
        val = Fraction(math.factorial(n1 + n2), n1)
        poly.add_keyval(key, val)
        key = (0, n1, -1, 1, 0, n1 + n2)
        poly.add_keyval(key, -val)
    if n1 == 0 and n2 >= 1:
        key = (0, 0, 0, 0, 0, n2)
        val = Fraction(math.factorial(n2), 1)
        poly.add_keyval(key, val)
    if n1 == 0 and n2 == 0:
        poly.add_keyval((0, 0, 0, 0, 0, 0), Fraction(1, 1))
    # multiply with M_J^*(n3)
    kf.extend(['mu_s', 'sigma_s'])
    poln = Poly()
    poln.set_keyfor(kf)
    pol1 = mnorm(n3)  # kf = ['mu_s', 'sigma_s^2']
    for k, v in poly.items():
        for K, V in pol1.items():
            poln.add_keyval(k + (K[0], 2 * K[1]), v * V)
    return poln

def d1_times_key(key, wrt):
    # key = (i, (n11, n21, n31, m1), ..., (n1l, n2l, n3l, ml))
    # multiply with [lambda (t-s)]
    knw = list(key)
    knw[0] += 1
    # multiply with \partial M_{E\!J,J,J^*} / \partial a_o (o = 1, 2, 3)
    if wrt == 1:
        n1n2n3 = (1, 0, 0)
    elif wrt == 2:
        n1n2n3 = (0, 1, 0)
    else:
        n1n2n3 = (0, 0, 1)
    #
    found = False
    for i in range(1, len(knw)):
        if knw[i][0:3] == n1n2n3:
            knw[i] = n1n2n3 + (knw[i][3] + 1,)
            found = True
            break
    if not found:
        knw.insert(1, n1n2n3 + (1,))
    return tuple(knw)

def dterm(key, coef, wrt):
    poly = Poly()
    # key = (i, (n11, n21, n31, m1), ..., (n1l, n2l, n3l, ml))
    for j in range(1, len(key)):
        knw = list(key)
        n1, n2, n3, m = key[j]
        coef_new = coef * m  # update coef
        # update exponent M^{(n1,n2,n3)m} -> M^{(n1,n2,n3)(m-1)}
        knw[j] = (n1, n2, n3, m - 1)
        # M^{(n1+1, n2, n3)1} or M^{(n1, n2+1, n3)1} or M^{(n1, n2, n3+1)1}
        if wrt == 1:
            n1 += 1
        elif wrt == 2:
            n2 += 1
        else:
            n3 += 1
        found = False
        for i in range(1, len(knw)):
            if knw[i][0:3] == (n1, n2, n3):
                knw[i] = knw[i][0:3] + (knw[i][3] + 1,)
                found = True
                break
        if not found:
            knw.append((n1, n2, n3, 1))
        if knw[j][3] == 0: del knw[j]  # delete M^{(n1,n2,n3)0}
        poly.add_keyval(tuple(knw), coef_new)
    return poly

def dmgf_ieziziz(poly, wrt):
    # step 1
    poly_sum = Poly()
    for key in poly:
        knw = d1_times_key(key, wrt)
        poly_sum.add_keyval(knw, poly[key])
    # step 2
    for key in poly:
        poly_sum.merge(dterm(key, poly[key], wrt))
    return poly_sum

def decode(poly):
    # poly with kf = ['(i, (n11, n21, n31, m1), ..., (n1l, n2l, n3l, ml))']
    kf = ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    for k in poly:
        i = k[0]
        pol1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
        pol1.set_keyfor(kf)
        for j in range(1, len(k)):
            n1, n2, n3, m = k[j]
            pol1 = pol1 * (moment_ejjj(n1, n2, n3) ** m)
        for key in pol1:
            knw = list(key)
            knw[2] += i
            knw[4] += i
            val = poly[k] * pol1[key]
            poln.add_keyval(tuple(knw), val)
    poln.remove_zero()
    return poln

def moment_ieziziz(n1, n2, n3):
    """joint moment of :math:`E[IEZ_{s,t}^{n_1} IZ_{s,t}^{n_2} IZ_{s,t}^{*n_3}]`

    :param integer n1: order of :math:`IEZ_{s,t}`
    :param integer n2: order of :math:`IZ_{s,t}`
    :param integer n3: order of :math:`IZ_{*s,t}`
    :return: poly with ``keyfor`` = ('e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd',
      'mu_v', 'mu_s', 'sigma_s')
    :rtype: Poly
    """
    if n1 < 0 or n2 < 0 or n3 < 0:
        raise ValueError('n1 and n2 and n3 must be non-negative integers!')
    kf0 = ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    M = {}
    kf = ['(i, (n11, n21, n31, m1), ..., (n1l, n2l, n3l, ml))']
    # n1 = n2 = n3 = 0
    poly = Poly({(0,): Fraction(1, 1)})
    poly.set_keyfor(kf)
    M[(0, 0, 0)] =poly
    #
    if n1 == 0 and n2 == 0 and n3 == 0:
        poln = Poly({(0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
        poln.set_keyfor(kf0)
        return poln
    # n1 + n2 + n3 = 1
    poly = Poly({(1, (1, 0, 0, 1)): Fraction(1, 1)})
    poly.set_keyfor(kf)
    M[(1, 0, 0)] = poly
    #
    poly = Poly({(1, (0, 1, 0, 1)): Fraction(1, 1)})
    poly.set_keyfor(kf)
    M[(0, 1, 0)] = poly
    #
    poly = Poly({(1, (0, 0, 1, 1)): Fraction(1, 1)})
    poly.set_keyfor(kf)
    M[(0, 0, 1)] = poly
    #
    if n1 == 1 and n2 == 0 and n3 == 0:
        poln = Poly({
            (1, 0, 0, 1, 1, 1, 0, 0): Fraction(1, 1),
            (0, 1, 0, 1, 1, 1, 0, 0): Fraction(-1, 1)
        })
        poln.set_keyfor(kf0)
        return poln
    if n1 == 0 and n2 == 1 and n3 == 0:
        poln = Poly({(0, 0, 1, 0, 1, 1, 0, 0): Fraction(1, 1)})
        poln.set_keyfor(kf0)
        return poln
    if n1 == 0 and n2 == 0 and n3 == 1:
        poln = Poly({(0, 0, 1, 0, 1, 0, 1, 0): Fraction(1, 1)})
        poln.set_keyfor(kf0)
        return poln
    # n1 + n2 + n3 > 1
    for n in range(2, n1 + n2 + n3):
        for i in range(n + 1):
            for j in range(n - i + 1):
                # (i, j, n-i-j)
                if i > 0 and j == 0:
                    M[(i, j, n-i-j)] = dmgf_ieziziz(M[(i-1, j, n-i-j)], 1)
                elif i == 0 and j > 0:
                    M[(i, j, n-i-j)] = dmgf_ieziziz(M[(i, j-1, n-i-j)], 2)
                elif i == 0 and j == 0:
                    M[(i, j, n-i-j)] = dmgf_ieziziz(M[(i, j, n-i-j-1)], 3)
                else:  # n1 > 0 and n2 > 0
                    M[(i, j, n-i-j)] = dmgf_ieziziz(M[(i, j-1, n-i-j)], 2)
    if n1 > 0 and n2 == 0:
        poln = dmgf_ieziziz(M[(n1-1, n2, n3)], 1)
    elif n1 == 0 and n2 > 0:
        poln = dmgf_ieziziz(M[(n1, n2-1, n3)], 2)
    elif n1 == 0 and n2 == 0:
        poln = dmgf_ieziziz(M[(n1, n2, n3-1)], 3)
    else:  # n1 > 0 and n2 > 0
        poln = dmgf_ieziziz(M[(n1, n2-1, n3)], 2)
    return decode(poln)

def poly2num(poly, par):
    # kf = ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v', 'mu_s', 'sigma_s']
    # s = 0
    k, h, lmbd = par['k'], par['h'], par['lmbd']
    mu_v, mu_s, sigma_s = par['mu_v'], par['mu_s'], par['sigma_s']
    #
    f = 0
    for K, V in poly.items():
        val = math.exp(K[0] * k * h) * (h ** K[2]) / (k ** K[3])
        val *= (lmbd ** K[4]) * (mu_v ** K[5]) * (mu_s ** K[6])
        val *= sigma_s ** K[7]
        f += val * V
    return f

def m_ieziziz(order, par):
    n1, n2, n3 = order
    moment = moment_ieziziz(n1, n2, n3)
    value = poly2num(moment, par)
    return value

if __name__ == "__main__":
    from pprint import pprint

    n = 4
    for n1 in range(n):
        for n2 in range(n):
            for n3 in range(n):
                poly = moment_ieziziz(n1, n2, n3)
                print(f"moment_ieziziz({n1}, {n2}, {n3}) = ")
                pprint(poly)
                print(f"which is a poly with keyfor = {poly.keyfor}\n")



"""
Covariance for One-Factor SV
"""
from ajdmom.poly import Poly
from ajdmom.utils import comb
from ajdmom.mdl_1fsv.cmom import ve_IEII
from ajdmom.mdl_1fsv.mom import (
    b_n,
    sub_v,
    moment_y,
    dpoly,
    poly2num
)


def exp_vn(l):
    """Expand :math:`v_n^l`

    :param int l: power
    :return: poly with attribute ``keyfor`` =
       ('e^{-knh}IE_n', 'e^{-kh}', 'v_{n-1}', 'theta', 'sigma_v').
    :rtype: Poly
    """
    polv = Poly()
    kf = ['e^{-knh}IE_n', 'e^{-kh}', 'v_{n-1}', 'theta', 'sigma_v']
    polv.set_keyfor(kf)
    for i in range(l, -1, -1):  # e^{-kh}v_{n-1}
        for j in range(l - i, -1, -1):  # 1*theta
            for p in range(l - i - j, -1, -1):  # -e^{-kh}*theta
                q = l - i - j - p  # sigma_v e^{-knh}IE_n
                key = (q, i + p, i, j + p, q)
                val = comb(l, [i, j, p, q]) * ((-1) ** p)
                polv.add_keyval(key, val)
    return polv


def ve_IEII_vn(n2, n3, n4, n5):
    """Compute :math:`v_n^{n_2}e^{-n_3k(n+1)h} IEII` and expand :math:`v_n`

    :param int n2: :math:`n_2` in above equation.
    :param int n3: :math:`n_3` in :math:`IEII` which is
       :math:`E[IE_{n+1}^{n_3}I_{n+1}^{n_4}I_{n+1}^{*n_5}|v_n]`.
    :param int n4: :math:`n_4` in :math:`IEII`.
    :param int n5: :math:`n_5` in :math:`IEII`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-knh}IE_n','e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v').
    :rtype: Poly
    """
    poly = ve_IEII(n2, n3, n4, n5)
    # ['e^{-kh}','h','v_n','k^{-}','theta','sigma_v']
    #
    # expand v_n
    poly_eIv = Poly()
    kf = ['e^{-knh}IE_n', 'e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'theta', 'sigma_v']
    poly_eIv.set_keyfor(kf)
    #
    for k1 in poly:
        polv = exp_vn(k1[2])
        # ['e^{-knh}IE_n', 'e^{-kh}', 'v_{n-1}', 'theta', 'sigma_v']
        for k2 in polv:
            key = (k2[0], k1[0] + k2[1], k1[1], k2[2], k1[3], k1[4] + k2[3], k1[5] + k2[4])
            val = poly[k1] * polv[k2]
            poly_eIv.add_keyval(key, val)
    return poly_eIv


def moment_inner_comb(l1, m1, m2, m3, m4, m5, poly_eIv):
    """Moment for this inner combination in expansion of :math:`y_n^{l_1}`

    :param int l1: *l1* in :math:`y_n^{l_1}`.
    :param int m1: times of the constant in :math:`y_n` being selected.
    :param int m2: times of :math:`v_{n-1}` being selected.
    :param int m3: times of :math:`IE_{n}` being selected.
    :param int m4: times of :math:`I_{n}` being selected.
    :param int m5: times of :math:`I_{n}^{*}` being selected.
    :param Poly poly_eIv: poly with attribute ``keyfor`` =
       ('e^{-knh}IE_n','e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v').
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','mu','theta','sigma_v','rho',
       'sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'theta', 'sigma_v']
    poly.set_keyfor(kf)
    #
    # combine and compute
    # k1: ['e^{-knh}IE_n','e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    # k2: ['e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    for k1 in poly_eIv:
        poln = ve_IEII(m2 + k1[3], m3 + k1[0], m4, m5)
        for k2 in poln:
            key = (k1[1] + k2[0], k1[2] + k2[1], k2[2], k1[4] + k2[3], k1[5] + k2[4], k1[6] + k2[5])
            val = poly_eIv[k1] * poln[k2]
            poly.add_keyval(key, val)
    #
    b = b_n(m1, m2, m3, m4, m5)
    # ['e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)']
    #
    keyfor = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho',
              'sqrt(1-rho^2)']
    keyIndexes = ([0, 1, 2, 3, -1, 4, 5, -1, -1], [0, 1, -1, 2, 3, 4, 5, 6, 7])
    poly = poly.mul_poly(b, keyIndexes, keyfor)
    #
    c = comb(l1, [m1, m2, m3, m4, m5])
    return c * poly


def moment_outer_comb(l2, n1, n2, n3, n4, n5, l1):
    r"""Moment for this outer combination in expansion of :math:`y_{n+1}^{l_2}`

    :param int l2: *l2* in :math:`y_{n+1}^{l_2}`.
    :param int n1: times of :math:`\theta` being selected.
    :param int n2: times of :math:`v_{n}` being selected.
    :param int n3: times of :math:`IE_{n+1}` being selected.
    :param int n4: times of :math:`I_{n+1}` being selected.
    :param int n5: times of :math:`I_{n+1}^{*}` being selected.
    :param int l1: *l1* in :math:`y_{n}^{l_1}`.
    :return: poly with attibute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','mu','theta','sigma_v','rho',
       'sqrt(1-rho^2)').
    :rtype: Poly
    """
    # c = c_n(l2, n1, n2, n3, n4, n5)
    c = comb(l2, [n1, n2, n3, n4, n5])
    b = b_n(n1, n2, n3, n4, n5)
    # ['e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)']
    #
    poly_eIv = ve_IEII_vn(n2, n3, n4, n5)
    # ['e^{-knh}IE_n','e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    #
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho',
          'sqrt(1-rho^2)']
    poly.set_keyfor(kf)
    #
    for m1 in range(l1, -1, -1):
        for m2 in range(l1 - m1, -1, -1):
            for m3 in range(l1 - m1 - m2, -1, -1):
                for m4 in range(l1 - m1 - m2 - m3, -1, -1):
                    m5 = l1 - m1 - m2 - m3 - m4
                    poly.merge(moment_inner_comb(l1, m1, m2, m3, m4, m5, poly_eIv))
    keyIndexes = ([0, 1, 2, 3, 4, 5, 6, 7, 8], [0, 1, -1, 2, 3, 4, 5, 6, 7])
    poly = poly.mul_poly(b, keyIndexes, kf)
    return c * poly


def moment_yy(l1, l2):
    """Moment :math:`E[y_n^{l_1}y_{n+1}^{l_2}]` as in equation :eq:`moment_yy`

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho',
          'sqrt(1-rho^2)']
    poly.set_keyfor(kf)
    #
    for n1 in range(l2, -1, -1):
        for n2 in range(l2 - n1, -1, -1):
            for n3 in range(l2 - n1 - n2, -1, -1):
                for n4 in range(l2 - n1 - n2 - n3, -1, -1):
                    n5 = l2 - n1 - n2 - n3 - n4
                    poly.merge(moment_outer_comb(l2, n1, n2, n3, n4, n5, l1))
    poly = sub_v(poly)
    poly.remove_zero()
    return poly


def cov_yy(l1, l2):
    """Covariance :math:`cov(y_n^{l_1},y_{n+1}^{l_2})` in equation :eq:`cov_yy`

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    #
    # moment_yy(l1,l2), moment_y(l1), moment_y(l2)
    # e^{-kh}, h, k^{-1}, mu, theta,  sigma_v, rho,  sqrt(1-rho^2)
    cov = moment_yy(l1, l2) - (moment_y(l1) * moment_y(l2))
    cov.remove_zero()
    return cov


##########
# scalar and (partial) derivative
##########

def cov(l1, l2, par):
    """Covariance in scalar

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :return: scalar of the covariance.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    value = poly2num(covariance, par)
    return value


def dcov(l1, l2, par, wrt):
    """Partial derivative of covariance w.r.t. parameter wrt

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    pold = dpoly(covariance, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsv.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l1 = 1 if len(args) == 0 else int(args[0])
    l2 = 1 if len(args) <= 1 else int(args[1])
    cov_l1_l2 = cov_yy(l1, l2)
    print(f"cov_yy({l1},{l2}) = \n")
    pprint(cov_l1_l2)
    print(f"\nwhich is a Poly with attribute keyfor =\n{cov_l1_l2.keyfor}")



"""
Conditional Central Moments for the SVCJ model, given :math:`v_0`
and the realized jumps in the variance.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.utils import cmnorm, comb, dbfactorial, fZ
from ajdmom.mdl_svvj.cond2_cmom import cmoments_y_to as cmoments_y_to_svvj


def cmoment_IZs(m, J):
    r"""conditional central moment :math:`\mathbb{E}[(\overline{IZ_t^{s}})^{m}
    |z^v(u), 0\le u \le t]`

    :param int m: power of :math:`\overline{IZ^s}_t`
    :param tuple J: jump sizes in the variance
    :return: poly with attribute ``keyfor`` = ('sigma_s'), noting that key is
      still a tuple (of one element)
    :rtype: Poly
    """
    kf = ['sigma_s']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    # special cases
    #
    if m == 0:  # return 1
        poly.add_keyval((0,), Frac(1, 1))
        return (poly)
    if len(J) == 0:  # return 0
        poly.add_keyval((0,), Frac(0, 1))
        return (poly)
    #
    # typical cases
    #
    if len(J) == 1:  # single normal variable
        poly = cmnorm(m)
        poly.set_keyfor(kf)
        return (poly)
    if len(J) == 2:
        for i in range(0, m + 1):
            c = math.comb(m, i)
            pol1 = cmnorm(i)
            pol2 = cmnorm(m - i)  # cmnorm(0) supported
            poln = pol1 * pol2
            poln.set_keyfor(kf)
            poly.merge(c * poln)
        return (poly)
    #
    # len(J) > 2
    #
    stack = []
    nextstep = "forward"
    while len(stack) < len(J):
        if nextstep == "forward":
            if len(stack) < len(J) - 2:
                stack.append(0)
                nextstep = "forward"
            else:
                left = m - sum(stack)
                for i in range(0, left + 1):
                    stack.extend([i, left - i])
                    index = stack.copy()
                    #
                    c = comb(m, index)
                    poln = Poly({(0,): Frac(1, 1)})
                    poln.set_keyfor(['sigma'])
                    for j in range(len(J)):
                        poln *= cmnorm(index[j])
                    poln.set_keyfor(['sigma_s'])
                    poly += c * poln
                    #
                    stack.pop()
                    stack.pop()
                # checking whether all indexes traversed
                if stack[0] == m: break
                #
                nextstep = "backward"
        else:
            left = m - sum(stack)
            if left == 0:
                stack.pop()
                nextstep = "backward"
            else:
                stack[-1] += 1
                nextstep = "forward"
    return poly


def cmoments_y_to(l, J):
    """conditional central moments of :math:`y_t` of orders :math:`0:l`

    :param int l: highest order of the conditional central moments to derive, >= 1
    :param tuple J: jump sizes in the variance
    :return: a list of polys with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'sigma_s')
    :rtype: list
    """
    cmoms = []
    cmoms_svvj = cmoments_y_to_svvj(l)  # 0-th to l-th central moments
    #
    kf = cmoms_svvj[0].keyfor + ('sigma_s',)
    #
    # special case: 0-th central moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, (), (), 0, 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    cmoms.append(P1)  # equiv to constant 1
    #
    # special case: 1-th central moment
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, (), (), 0, 0, 0, 0): Frac(0, 1)})
    P0.set_keyfor(kf)
    cmoms.append(P0)  # equiv to constant 0
    #
    # typical cases
    #
    for n in range(2, l + 1):
        poly = Poly()
        poly.set_keyfor(kf)
        for n1 in range(n, -1, -1):
            c = math.comb(n, n1)
            pol1 = cmoms_svvj[n1]  # 0-th central moment supported
            pol2 = cmoment_IZs(n - n1, J)  # kf = ['sigma_s']
            for k1 in pol1:
                for k2 in pol2:
                    key = k1 + k2
                    poly.add_keyval(key, c * pol1[k1] * pol2[k2])
        poly.remove_zero()
        cmoms.append(poly)
    return cmoms


def cmoment_y(l, J):
    cmoms = cmoments_y_to(l, J)
    return cmoms[-1]


##########
# scalar
##########
def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attibute ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'sigma_s')
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    rho = par['rho']
    h = par['h']  # t = h
    J = par['jumpsize']  # list of jump sizes
    s = par['jumptime']  # list of jump time points
    sigma_s = par['sigma_s']
    #
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= ((v0 - theta) ** K[3]) * (theta ** K[4]) * (sigma ** K[5])
        #
        l, o= K[6], K[7]
        val *= fZ(l, o, k, s, J)
        #
        val *= (sigma / (2 * k)) ** K[8]
        val *= (rho - sigma / (2 * k)) ** K[9]
        val *= (1 - rho ** 2) ** (K[10] / 2)
        #
        val *= sigma_s ** K[11]
        #
        f += val * poly[K]
    return f


def cm(l, par):
    """conditional central moment in scalar

    :param int l: order of the conditional central moment
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the conditional central moment
    :rtype: float
    """
    J = par['jumpsize']  # list of jump sizes
    cmoments = cmoments_y_to(l, J)
    cmoment = cmoments[-1]
    value = poly2num(cmoment, par)
    return value


if __name__ == "__main__":
    from pprint import pprint
    import sys
    import random

    print('Example usage of the module function\n')
    numJ = 5  # special cases: 0, 1, 2
    jumpsize = [random.expovariate(lambd=1.0) for i in range(numJ)]
    args = sys.argv[1:]
    l = 2 if len(args) == 0 else int(args[0])
    cmoms = cmoments_y_to(l, jumpsize)  # 0-th to l-th central moments
    kf = cmoms[0].keyfor
    print(f"cmoment_y({l}, J) = \n")
    pprint(cmoms[l])
    print(f"\nwhich is a poly with keyfor = \n{kf}")



"""
Covariance for the SVVJ model
"""
import math
from fractions import Fraction

from ajdmom import Poly
from ajdmom.utils import comb, simplify_rho
from ajdmom.mdl_svcj.cond_ieii_ieziziz_mom import moment_ieii_ieziziz
from ajdmom.mdl_svcj.mom import moment_v_theta
from ajdmom.mdl_svvj.mom import key_times_poly, poly2num, moment_y, expand_Ev


def exp_vn_theta(l):
    r"""Expand :math:`(v_n - \theta)^l`

    :param int l: power of :math:`(v_n - \theta)`
    :return: poly with attribute ``keyfor`` =
      ('e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'v_{n-1}-theta', 'sigma_v')
    :rtype: Poly
    """
    polv = Poly()
    kf = ['e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'v_{n-1}-theta', 'sigma_v']
    polv.set_keyfor(kf)
    for i in range(l + 1):
        for j in range(l - i + 1):
            p = l - i - j
            bino = comb(l, [i, j, p])
            key = (j, p, i, i, j)
            polv.add_keyval(key, Fraction(bino, 1))
    return polv

def ve_IEII_IEZIZ(n0, n1, n2, n3, n4, n5):
    r"""
    Derive :math:`(v_n-\theta)^{n_0} e^{-(n_1+n_4)k(n+1)h}
    \mathbb{E}[I\!E_{n+1}^{n_1} I_{n+1}^{n_2} I_{n+1}^{*n_3}
    I\!E\!Z_{n+1}^{n_4} I\!Z_{n+1}^{n_5}]`
    """
    ve_poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'v_n-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
    ve_poly.set_keyfor(kf)
    #
    poly = moment_ieii_ieziziz(n1, n2, n3, n4, n5, 0)
    # with keyfor
    # 'e^{knh}' + ('e^{kh}', 'h', 'k^{-}', 'v_n-theta', 'theta', 'sigma', 'lmbd', 'mu_v',
    # 'mu_s', 'sigma_s')
    # and the power for the eliminated 'e^{knh}' always = m1 + m4
    # may error: further check!
    for k, v in poly.items():
        # × e^{-(n_1+n_4)k(n+1)h}          × (v_n-theta)^{m_0}
        key = (n1 + n4 - k[0], k[1], k[2], n0 + k[3], k[4], k[5], k[6], k[7])
        ve_poly.add_keyval(key, v)
    return ve_poly

def ve_IEII_IEZIZ_vn(n0, n1, n2, n3, n4, n5):
    r"""Derive :math:`(v_n-\theta)^{n_0}e^{-(n_1+n_4)k(n+1)h} IEII_IEZIZ` and expand :math:`v_n-\theta`

    :param int n0: order of :math:`v_n-\theta`
    :param int n1: order of :math:`IE_{n+1}`
    :param int n2: order of :math:`I_{n+1}`
    :param int n3: order of :math:`I_{n+1}^{*}`
    :param int n4: order of :math:`IEZ_{n+1}`
    :param int n5: order of :math:`IZ_{n+1}`
    :return: poly with attribute ``keyfor`` =
      ('e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta',
      'theta', 'sigma_v', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    ve_poly = ve_IEII_IEZIZ(n0, n1, n2, n3, n4, n5)
    #
    # expand (v_n - theta)
    #
    poly = Poly()
    kf = ['e^{-knh}IE_n', 'e^{-knh}IEZ_n',
          'e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
    poly.set_keyfor(kf)
    #
    for k1 in ve_poly:
        # ['e^{-kh}', 'h', 'k^{-}', 'v_n-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
        polv = exp_vn_theta(k1[3])
        # ['e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'v_{n-1}-theta', 'sigma_v']
        for k2 in polv:
            key = (        k2[0],  # 'e^{-knh}IE_n'
                           k2[1],  # 'e^{-knh}IEZ_n'
                   k1[0] + k2[2],  # 'e^{-kh}'
                   k1[1],          # 'h'
                   k1[2],          # 'k^{-}'
                           k2[3],  # 'v_{n-1}-theta'
                   k1[4],          # 'theta'
                   k1[5] + k2[4],  # 'sigma_v'
                   k1[6],          # 'lmbd'
                   k1[7])          # 'mu_v'
            val = ve_poly[k1] * polv[k2]
            poly.add_keyval(key, val)
    return poly

def b_n(m0, m1, m2, m3, m4, m5, m6):
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v',
          'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(m0 + 1):         # (1 - e^{-kh})/(2k)
        for j in range(m2 + 1):     # (rho - sigma_v/2k)
            for o in range(m6 + 1): # (mu - theta/2)h
                bino = math.comb(m0, i) * math.comb(m2, j) * math.comb(m6, o)
                sign = (-1) ** (m0 + m5 + i + j + o)
                key = (i,       # 'e^{-kh}'
                       m6,      # 'h'
                       m0 + m1 + m4 + m5 + j, # 'k^{-}'
                       m6 - o,  # 'mu'
                       o,       # theta
                       m1 + j,  # sigma_v
                       m2 - j,  # rho
                       m3,      # sqrt(1-rho^2)
                       0, 0)
                val = sign * bino * Fraction(1, 2 ** (m0 + m1 + m4 + m5 + j + o))
                poly.add_keyval(key, val)
    return poly

def moment_inner_comb(l1, m0, m1, m2, m3, m4, m5, m6, poly_eIv):
    r"""Moment for this inner combination in expansion of :math:`y_n^{l_1}`

    :param int l1: order in :math:`y_n^{l_1}`.
    :param int m0: times of :math:`(v_{n-1}-\theta)` being selected.
    :param int m1: times of :math:`I\!E_n` being selected.
    :param int m2: times of :math:`I_n` being selected.
    :param int m3: times of :math:`I_n^{*}` being selected.
    :param int m4: times of :math:`I\!E\!Z_n` being selected.
    :param int m5: times of :math:`I\!Z_n` being selected.
    :param int m6: times of :math:`(\mu - \theta/2)h` being selected.
    :param Poly poly_eIv: poly with attribute ``keyfor`` =
      ('e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta',
      'theta', 'sigma_v', 'lmbd', 'mu_v')
    :return: poly with attribute ``keyfor`` =
      (e^{-kh}', 'h', 'k^{-}', 'mu', 'v_{n-1}-theta',
      'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
    poly.set_keyfor(kf)
    #
    # combine and compute
    # k1: ['e^{-knh}IE_n', 'e^{-knh}IEZ_n', 'e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta',
    #      'theta', 'sigma_v', 'lmbd', 'mu_v']
    # k2: ['e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
    for k1 in poly_eIv:
        poln = ve_IEII_IEZIZ(m0 + k1[5], m1 + k1[0], m2, m3, m4 + k1[1], m5)
        for k2 in poln:
            key = (k1[2] + k2[0],  # 'e^{-kh}'
                   k1[3] + k2[1],  # 'h'
                   k1[4] + k2[2],  # k^{-}
                   # k1[5] + k2[3],  # v_{n-1}-theta
                           k2[3],  # v_{n-1}-theta
                   k1[6] + k2[4],  # theta
                   k1[7] + k2[5],  # sigma_v
                   k1[8] + k2[6],  # lmbd
                   k1[9] + k2[7])  # mu_v
            val = poly_eIv[k1] * poln[k2]
            poly.add_keyval(key, val)
    b = b_n(m0, m1, m2, m3, m4, m5, m6)
    # kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v',
    #       'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    keyfor = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'v_{n-1}-theta', 'theta', 'sigma_v',
              'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    keyIndexes = (
        [0, 1, 2, -1, 3, 4, 5, -1, -1, 6, 7],  # poly index
        [0, 1, 2, 3, -1, 4, 5, 6, 7, 8, 9]     # b index
    )
    poly = poly.mul_poly(b, keyIndexes, keyfor)
    c = comb(l1, [m0, m1, m2, m3, m4, m5, m6])
    return c * poly

def moment_outer_comb(l2, n0, n1, n2, n3, n4, n5, n6, l1):
    """Moment for this outer combination in expansion of :math:`y_{n+1}^{l_2}`"""
    c = comb(l2, [n0, n1, n2, n3, n4, n5, n6])
    b = b_n(n0, n1, n2, n3, n4, n5, n6)
    # kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v',
    #       'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    poly_eIv = ve_IEII_IEZIZ_vn(n0, n1, n2, n3, n4, n5)
    # kf = ['e^{-knh}IE_n', 'e^{-knh}IEZ_n',
    #       'e^{-kh}', 'h', 'k^{-}', 'v_{n-1}-theta', 'theta', 'sigma_v', 'lmbd', 'mu_v']
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'v_{n-1}-theta', 'theta', 'sigma_v',
          'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    poly.set_keyfor(kf)
    #
    for m0 in range(l1 + 1):
        for m1 in range(l1 - m0 + 1):
            for m2 in range(l1 - m0 - m1 + 1):
                for m3 in range(l1 - m0 - m1 - m2 + 1):
                    for m4 in range(l1 - m0 - m1 - m2 - m3 + 1):
                        for m5 in range(l1 - m0 - m1 - m2 - m3 - m4 + 1):
                            m6 = l1 - m0 - m1 - m2 - m3 - m4 - m5
                            poly.merge(moment_inner_comb(l1, m0, m1, m2, m3, m4, m5, m6, poly_eIv))
    keyIndexes = (
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],  # poly index
        [0, 1, 2, 3, -1, 4, 5, 6, 7, 8, 9]   # b index
    )
    poly = poly.mul_poly(b, keyIndexes, kf)
    return c * poly

def moment_yy(l1, l2):
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'v_{n-1}-theta', 'theta', 'sigma_v',
          'rho', 'sqrt(1-rho^2)', 'lmbd', 'mu_v']
    poly.set_keyfor(kf)
    #
    for n0 in range(l2 + 1):
        for n1 in range(l2 - n0 + 1):
            for n2 in range(l2 - n0 - n1 + 1):
                for n3 in range(l2 - n0 - n1 - n2 + 1):
                    for n4 in range(l2 - n0 - n1 - n2 - n3 + 1):
                        for n5 in range(l2 - n0 - n1 - n2 - n3 - n4 + 1):
                            n6 =l2 - n0 - n1 - n2 - n3 - n4 - n5
                            poly.merge(moment_outer_comb(l2, n0, n1, n2, n3, n4, n5, n6, l1))
    poly = simplify_rho(poly, 8)
    # ['e^{-kh}', 'h', 'k^{-}', 'mu', 'v_{n-1}-theta', 'theta', 'sigma_v', 'rho',
    #  'lmbd', 'mu_v']
    #
    # be careful of the changes, 'e^{-kh}' -> 'e^{kt}', 'h' -> 't', 'sigma_v' -> 'sigma'
    #
    kf = ['e^{kt}', 't', 'k^{-}', 'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        k = (-k[0],) + k[1:]         # from 'e^{-kh}' to 'e^{kt}'
        # k: ['e^{kt}', 't', 'k^{-}', 'mu', 'v0-theta', 'theta', 'sigma', 'rho',
        #      'lmbd', 'mu_v']
        pol1 = moment_v_theta(k[4])  # ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
        pol2 = key_times_poly(v, k, pol1)
        poln.merge(pol2)
    return poln


def cov_yy(l1, l2):
    # moment_yy: ['e^{kt}', 't', 'k^{-}', 'mu', 'E[v]', 'theta', 'sigma', 'rho', 'lmbd', 'mu_v']
    # moment_y:  ['e^{kt}', 't', 'k^{-}', 'mu', 'E[v]', 'theta', 'sigma', 'rho', 'lmbd', 'mu_v']
    cov = moment_yy(l1, l2) - (moment_y(l1) * moment_y(l2))
    cov.remove_zero()
    return cov

def cov(l1, l2, par):
    covariance = cov_yy(l1, l2)
    value = poly2num(covariance, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_svvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l1 = 2 if len(args) == 0 else int(args[0])
    l2 = 1 if len(args) <= 1 else int(args[1])
    cov_l1_l2 = cov_yy(l1, l2)
    cov_l1_l2 = expand_Ev(cov_l1_l2)
    print(f"cov_yy({l1},{l2}) = ")
    pprint(cov_l1_l2)
    print(f"which is a Poly with attribute keyfor =\n{cov_l1_l2.keyfor}")
    print(f'total {len(cov_l1_l2)} key-values\n')

    cov12 = expand_Ev(cov_yy(1, 2))
    cov21 = expand_Ev(cov_yy(2, 1))
    diff = cov12 - cov21
    pprint(diff)
    print(f'cov12 - cov21: total {len(diff)} key-values')



"""
Conditional Moments for the SVVJ model, given the initial state of
the variance and the realized jumps in the variance.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.utils import comb, fZ
from ajdmom.mdl_svvj.cond2_cmom import cmoments_y_to


def moment_comb(n, n1, n2, n3, n4, PY):
    r"""moment for this combination in the expansion of :math:`y_t^n`

    :param int n: n in :math:`E[y_t^n]`
    :param int n1: power of :math:`(\mu -\theta/2)t`
    :param int n2: power of :math:`(v_0 - \theta)\beta_t`
    :param int n3: power of :math:`I\!E\!Z_t`
    :param int n4: power of :math:`I\!Z_t`
    :param Poly PY: poly of :math:`E[\overline{y}_t^{n_5}|v_0, z_s, 0\le s\le t]`
    :return: poly with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)'), now 'p' encoding power
      of :math:`I\!Z_t`.
    :rtype: Poly
    """
    kf = PY.keyfor[0:3] + ('beta_t', 'mu-theta/2') + PY.keyfor[3:8] + ('p',)
    kf += PY.keyfor[8:]
    poly = Poly()
    poly.set_keyfor(kf)
    #
    n5 = n - n1 - n2 - n3 - n4
    c = comb(n, [n1, n2, n3, n4, n5])
    c *= ((-1) ** (n2 + n4)) * Frac(1, 2 ** (n3 + n4))
    #
    # PY.keyfor = ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
    #       'l_{1:n}', 'o_{1:n}',
    #       'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    for k in PY:
        key1 = (k[0] - n3, k[1] + n1, k[2] + n3 + n4, n2, n1, k[3] + n2, k[4], k[5])
        key3 = k[8:]
        #
        k6, k7 = k[6], k[7]
        # × IEZ_t^n3
        #
        # expand k6: e^{k(s_i1\vee s_in)}, k7: s_i
        #
        k6 += tuple(0 for _ in range(n3))
        k7 += tuple(0 for _ in range(n3))
        # × IZ_t^n4
        k8 = n4
        #
        key = key1 + (k6, k7, k8) + key3
        poly.add_keyval(key, c * PY[k])
    return poly


def moments_y_to(l):
    """conditional moments of :math:`y_t` with orders :math:`0:l`

    :param int l: highest order of the conditional moments to derive, l >= 1
    :return: a list of polys with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :rtype: Poly
    """
    moms = []
    cmoms = cmoments_y_to(l)  # get 0:l-th central moments at once
    #
    kf = cmoms[0].keyfor[0:3] + ('beta_t', 'mu-theta/2') + cmoms[0].keyfor[3:8]
    kf += ('p',) + cmoms[0].keyfor[8:]
    #
    # special case: 0-th moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, (), (), 0, 0, 0, 0): Frac(1, 1)})
    #                                           p
    P1.set_keyfor(kf)
    moms.append(P1)  # equiv to constant 1
    #
    # typical cases
    #
    for n in range(1, l + 1):
        poly = Poly()
        poly.set_keyfor(kf)
        for n1 in range(n, -1, -1):
            for n2 in range(n - n1, -1, -1):
                for n3 in range(n - n1 - n2, -1, -1):
                    for n4 in range(n - n1 - n2 - n3, -1, -1):
                        n5 = n - n1 - n2 - n3 - n4
                        poly.merge(moment_comb(n, n1, n2, n3, n4, cmoms[n5]))
        poly.remove_zero()
        moms.append(poly)
    return moms


def moment_y(l):
    moments = moments_y_to(l)
    return moments[-1]


##########
# scalar
##########

def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    rho = par['rho']
    h = par['h']  # t= h
    J = par['jumpsize']  # vector of jump sizes
    s = par['jumptime']  # vector of jump time points
    #
    beta_t = (1 - math.exp(-k * h)) / (2 * k)
    #
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (beta_t ** K[3]) * ((mu - theta / 2) ** K[4]) * ((v0 - theta) ** K[5])
        val *= (theta ** K[6]) * (sigma ** K[7])
        #
        l, o, p = K[8], K[9], K[10]
        #
        val *= fZ(l, o, k, s, J)
        #
        IZ = sum(J)
        val *= IZ ** p
        #
        val *= ((sigma / (2 * k)) ** K[11]) * ((rho - sigma / (2 * k)) ** K[12])
        val *= ((1 - rho ** 2) ** (K[13] / 2))
        #
        f += val * poly[K]
    return f


def m(l, par):
    """conditional moment in scalar

    :param int l: order of the conditional moment
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the moment
    :rtype: float
    """
    moments = moments_y_to(l)
    moment = moments[-1]
    value = poly2num(moment, par)
    return value


if __name__ == "__main__":
    import sys
    from pprint import pprint

    print('Example usage of the module function\n')
    args = sys.argv[1:]
    l = 1 if len(args) == 0 else int(args[0])
    moms = moments_y_to(l)  # 0-th - l-th moments
    kf = moms[0].keyfor
    print(f"moment_y({l}) = ")
    pprint(moms[l])
    print(f"\nwhich is a poly with keyfor = \n{kf}")



"""
Covariances for Two-Factor SV
"""
from ajdmom.poly import Poly
from ajdmom.itos_mom import moment_IEI_IEII, t_mul_t0
from ajdmom.utils import comb
from ajdmom.mdl_1fsv.cov import exp_vn

from ajdmom.mdl_2fsv.cmom import vvee_IEI_IEII
from ajdmom.mdl_2fsv.mom import (
    b_n,
    sub_v,
    moment_y,
    dpoly,
    poly2num
)


def vvee_IEI_IEII_vnvn(n2, n3, n4, n5, n6, n7, n8):
    r"""Multiply :math:`IEI\_IEII` by :math:`v_{1,n}v_{2,n}e^{-k_1(n+1)h}e^{-k_2(n+1)h}`

    Compute
    :math:`E[v_{1,n}^{n_2}v_{2,n}^{n_3}e^{-n_4k_1(n+1)h}e^{-n_6k_2(n+1)h}
    eI_{1,n+1}^{n_4} I_{1,n+1}^{n_5} eI_{2,n+1}^{n_6} I_{2,n+1}^{n_7}
    I_{n+1}^{*n_8}|v_{1,n}v_{2,n}]` and expand :math:`v_{1,n}` and
    :math:`v_{2,n}`.

    :param int n2: times of :math:`v_{1,n}` being selected.
    :param int n3: times of :math:`v_{2,n}` being selected.
    :param int n4: times of :math:`eI_{1,n+1}` being selected.
    :param int n5: times of :math:`I_{1,n+1}` being selected.
    :param int n6: times of :math:`eI_{2,n+1}` being selected.
    :param int n7: times of :math:`I_{2,n+1}` being selected.
    :param int n8: times of :math:`I_{n+1}^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('e^{-k1*nh}eI_{1,n}','e^{-k2*nh}eI_{2,n}',
       '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = vvee_IEI_IEII(n2, n3, n4, n5, n6, n7, n8)
    # ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    #  'e^{-(n1*k1+n2*k2)h}','h',
    # 'v_{1,n}','theta1','sigma_v1', 'v_{2,n}','theta2','sigma_v2']
    #
    # expand v1_n v2_n
    poly_eIv = Poly()
    kf = ['e^{-k1*nh}eI_{1,n}', 'e^{-k2*nh}eI_{2,n}',
          '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly_eIv.set_keyfor(kf)
    #
    poln = Poly()
    for k in poly:
        polv = exp_vn(k[3])  # expand v1_n
        # ['e^{-k1nh}eI_{1,n}', 'e^{-k1h}', 'v_{1,n-1}', 'theta1', 'sigma_v1']
        for k1 in polv:
            key = (
                k1[0], 0,  # 'e^{-k1*nh}eI_{1,n}', 'e^{-k2*nh}eI_{2,n}'
                k[0],  # '(n_1m, n_2m, i_m), ..., (n_11, n_21, i_1)'
                (k[1][0] + k1[1], k[1][1]),  # 'e^{-(n1*k1+n2*k2)h}'
                k[2],  # 'h'
                0 + k1[2],  # 'v_{1,n-1}', k[3] has been expanded
                k[4] + k1[3],  # 'theta1'
                k[5] + k1[4],  # 'sigma_v1'
                k[6], k[7], k[8]  # 'v_{2,n}', 'theta2', 'sigma_v2'
            )
            val = poly[k] * polv[k1]
            poln.add_keyval(key, val)
    #
    for k in poln:
        polv = exp_vn(k[8])  # expand v2_n
        # ['e^{-k2nh}eI_{2,n}', 'e^{-k2h}', 'v_{2,n-1}', 'theta2', 'sigma_v2']
        for k1 in polv:
            key = (
                k[0], k1[0],  # 'e^{-k1*nh}eI_{1,n}', 'e^{-k2*nh}eI_{2,n}'
                k[2],  # '(n_1m, n_2m, i_m), ..., (n_11, n_21, i_1)'
                (k[3][0], k[3][1] + k1[1]),  # 'e^{-(n1*k1+n2*k2)h}'
                k[4],  # 'h'
                k[5], k[6], k[7],  # 'v_{1,n-1}', 'theta1', 'sigma_v1'
                0 + k1[2],  # 'v_{2,n-1}', k[8] has been expanded
                k[9] + k1[3],  # 'theta_2'
                k[10] + k1[4]  # 'sigma_v2'
            )
            val = poln[k] * polv[k1]
            poly_eIv.add_keyval(key, val)
    return poly_eIv


def moment_inner_comb(l1, m1, m2, m3, m4, m5, m6, m7, m8, poly_eIv):
    """Moment for this inner combination in expansion of :math:`y_n^{l_1}`

    :param int l1: *l1* in :math:`y_n^{l_1}`.
    :param int m1: times of constant being selected.
    :param int m2: times of :math:`v_{1,n-1}` being selected.
    :param int m3: times of :math:`v_{2,n-1}` being selected.
    :param int m4: times of :math:`eI_{1,n}` being selected.
    :param int m5: times of :math:`I_{1,n}` being selected.
    :param int m6: times of :math:`eI_{2,n}` being selected.
    :param int m7: times of :math:`I_{2,n}` being selected.
    :param int m8: times of :math:`I_{n}^{*}` being selected.
    :param Poly poly_eIv: poly with attribute ``keyfor`` =
       ('e^{-k1*nh}eI_{1,n}','e^{-k2*nh}eI_{2,n}',
       '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly.set_keyfor(kf)
    #
    # combine and compute
    #
    # k1: ['e^{-k1*nh}eI_{1,n}','e^{-k2*nh}eI_{2,n}',
    #
    # '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    # 'e^{-(n1*k1+n2*k2)h}','h', 'v_{1,n-1}','theta1','sigma_v1',
    # 'v_{2,n-1}','theta2','sigma_v2']
    #
    # k2: ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    # 'e^{-(n1*k1+n2*k2)h}','h', 'v_{1,n-1}','theta1','sigma_v1',
    # 'v_{2,n-1}','theta2','sigma_v2']
    #
    for k1 in poly_eIv:
        poln = vvee_IEI_IEII(m2 + k1[5], m3 + k1[8], m4 + k1[0], m5, m6 + k1[1], m7, m8)
        for k2 in poln:
            t0 = k1[2]  # simply: t0 = k1[2] + k2[0]        # first special one
            for t in k2[0]: t0 = t_mul_t0(t, t0)  # carefully treat
            t1 = (k1[3][0] + k2[1][0], k1[3][1] + k2[1][1])  # second special one
            #              h,           mu,
            key = (t0, t1, k2[2] + k1[4], 0,
                   # v_{1,n-1},  theta1,    sigma_v1,
                   k2[3], k2[4] + k1[6], k2[5] + k1[7],
                   # v_{2,n-1},  theta2,    sigma_v2
                   k2[6], k2[7] + k1[9], k2[8] + k1[10])
            val = poly_eIv[k1] * poln[k2]
            poly.add_keyval(key, val)
    #
    # multiply with coefficient
    #
    b = b_n(m1, m2, m3, m4, m5, m6, m7)
    # ['e^{-k1h}','k1^{-}','theta1','sigma_v1',
    #  'e^{-k2h}','k2^{-}','theta2','sigma_v2', 'h','mu']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k1 in poly:
        for k2 in b:
            # t0 = ((1,0,k2[1]), (0,1,k2[5])) + k1[0]
            t0 = t_mul_t0((1, 0, k2[1]), k1[0])
            t0 = t_mul_t0((0, 1, k2[5]), t0)
            #
            t1 = (k1[1][0] + k2[0], k1[1][1] + k2[4])
            #              h,           mu,
            key = (t0, t1, k1[2] + k2[8], k1[3] + k2[9],
                   # v_{1,n-1},  theta1,    sigma_v1,
                   k1[4], k1[5] + k2[2], k1[6] + k2[3],
                   # v_{2,n-1},  theta2,    sigma_v2,
                   k1[7], k1[8] + k2[6], k1[9] + k2[7])
            val = poly[k1] * b[k2]
            poln.add_keyval(key, val)
    #
    c = comb(l1, [m1, m2, m3, m4, m5, m6, m7, m8])
    return c * poln


def moment_outer_comb(l2, n1, n2, n3, n4, n5, n6, n7, n8, l1):
    """Moment for this outer combination in expansion of :math:`y_{n+1}^{l_2}`

    :param int l2: *l2* in :math:`y_{n+1}^{l_2}`.
    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{1,n}` being selected.
    :param int n3: times of :math:`v_{2,n}` being selected.
    :param int n4: times of :math:`eI_{1,n+1}` being selected.
    :param int n5: times of :math:`I_{1,n+1}` being selected.
    :param int n6: times of :math:`eI_{2,n+1}` being selected.
    :param int n7: times of :math:`I_{2,n+1}` being selected.
    :param int n8: times of :math:`I_{n+1}^{*}` being selected.
    :param int l1: *l1* in :math:`y_{n}^{l_1}`.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poly_eIv = vvee_IEI_IEII_vnvn(n2, n3, n4, n5, n6, n7, n8)
    #
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly.set_keyfor(kf)
    #
    for m1 in range(l1, -1, -1):
        for m2 in range(l1 - m1, -1, -1):
            for m3 in range(l1 - m1 - m2, -1, -1):
                for m4 in range(l1 - m1 - m2 - m3, -1, -1):
                    for m5 in range(l1 - m1 - m2 - m3 - m4, -1, -1):
                        for m6 in range(l1 - m1 - m2 - m3 - m4 - m5, -1, -1):
                            for m7 in range(l1 - m1 - m2 - m3 - m4 - m5 - m6, -1, -1):
                                m8 = l1 - m1 - m2 - m3 - m4 - m5 - m6 - m7
                                poln = moment_inner_comb(l1, m1, m2, m3, m4, m5, m6, m7, m8, poly_eIv)
                                poly.merge(poln)
    #
    b = b_n(n1, n2, n3, n4, n5, n6, n7)
    # ['e^{-k1h}','k1^{-}','theta1','sigma_v1',
    #  'e^{-k2h}','k2^{-}','theta2','sigma_v2', 'h','mu']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k1 in poly:
        for k2 in b:
            # t0 = ((1,0,k2[1]), (0,1,k2[5])) + k1[0]
            t0 = t_mul_t0((1, 0, k2[1]), k1[0])
            t0 = t_mul_t0((0, 1, k2[5]), t0)
            #
            t1 = (k1[1][0] + k2[0], k1[1][1] + k2[4])
            #              h,           mu,
            key = (t0, t1, k1[2] + k2[8], k1[3] + k2[9],
                   # v_{1,n-1},  theta1,    sigma_v1,
                   k1[4], k1[5] + k2[2], k1[6] + k2[3],
                   # v_{2,n-1},  theta2,    sigma_v2,
                   k1[7], k1[8] + k2[6], k1[9] + k2[7])
            val = poly[k1] * b[k2]
            poln.add_keyval(key, val)
    #
    c = comb(l2, [n1, n2, n3, n4, n5, n6, n7, n8])
    return c * poln


def moment_yy(l1, l2):
    """Co-Moment :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`

    :param int l1: *l1* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: *l2* in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly.set_keyfor(kf)
    #
    for n1 in range(l2, -1, -1):
        for n2 in range(l2 - n1, -1, -1):
            for n3 in range(l2 - n1 - n2, -1, -1):
                for n4 in range(l2 - n1 - n2 - n3, -1, -1):
                    for n5 in range(l2 - n1 - n2 - n3 - n4, -1, -1):
                        for n6 in range(l2 - n1 - n2 - n3 - n4 - n5, -1, -1):
                            for n7 in range(l2 - n1 - n2 - n3 - n4 - n5 - n6, -1, -1):
                                n8 = l2 - n1 - n2 - n3 - n4 - n5 - n6 - n7
                                poln = moment_outer_comb(l2, n1, n2, n3, n4, n5, n6, n7, n8, l1)
                                poly.merge(poln)
    poly_sum = sub_v(poly)
    poly_sum.remove_zero()
    return poly_sum


def cov_yy(l1, l2):
    """Covariance :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    # cov = moment_yy(l1, l2) - (moment_y(l1) * moment_y(l2))
    poln = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2']
    poln.set_keyfor(kf)
    #
    pol1 = moment_y(l1)  # keyfor = kf
    pol2 = moment_y(l2)  # keyfor = kf
    for k1 in pol1:
        for k2 in pol2:
            t0 = k1[0]  # simply: t0 = k1[0] + k2[0]    # first special one
            for t in k2[0]: t0 = t_mul_t0(t, t0)  # carefully treat
            t1 = (k1[1][0] + k2[1][0], k1[1][1] + k2[1][1])  # second special one
            #
            key = (t0, t1) + tuple(k1[i] + k2[i] for i in range(2, 8))
            val = pol1[k1] * pol2[k2]
            poln.add_keyval(key, val)
    #
    cov = moment_yy(l1, l2) - poln  # moment_yy(l1,l2) with keyfor = kf
    cov.remove_zero()
    return cov


##########
# scalar and (partial) derivative
##########

def cov(l1, l2, par):
    """Covariance in scalar

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :return: scalar of the covariance.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    value = poly2num(covariance, par)
    return value


def dcov(l1, l2, par, wrt):
    """Partial derivative of covariance w.r.t. parameter wrt

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    pold = dpoly(covariance, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_2fsv.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l1 = 1 if len(args) == 0 else int(args[0])
    l2 = 1 if len(args) <= 1 else int(args[1])
    cov_l1_l2 = cov_yy(l1, l2)
    print(f"cov_yy({l1},{l2}) = \n")
    pprint(cov_l1_l2)
    print(f"\nwhich is a Poly with attribute keyfor =\n{cov_l1_l2.keyfor}")



"""
Subpackage for Two-Factor :abbr:`SV(Stochastic Volatility)` model

Modules ``mom``, ``cmom``, ``cov``.
"""
from .cmom import cmoment_y
from .mom import moment_y
from .cov import cov_yy, moment_yy



"""
Class Poly, a customized dictionary, derived from :code:`UserDict`

I defined a new class :py:class:`~ajdmom.poly.Poly`
to extend the :class:`~collections.UserDict` class
from the Python Standard Library :code:`collections`, within which
I defined an attribute :py:attr:`~ajdmom.poly.Poly.keyfor` and the
following magic methods:

+-------------+---------------------+---------------------------------------+
|Magic Method | Example             |Function                               |
+=============+=====================+=======================================+
|__add__()    |:code:`poly1 + poly2`|Addition between two polynomials       |
+-------------+---------------------+---------------------------------------+
|__sub__()    |:code:`poly1 - poly2`|Subtraction between two polynomials    |
+-------------+---------------------+---------------------------------------+
|__mul__()    |:code:`poly1 * poly2`|Multiplication between two polynomials |
+-------------+---------------------+---------------------------------------+
|__rmul__()   |:code:`c * poly2`    |Reversely multiply by a constant       |
+-------------+---------------------+---------------------------------------+
|__pow__()    |:code:`poly ** 2`    |Raise a polynomial to  power n         |
+-------------+---------------------+---------------------------------------+

It should be noted that:

* Addition and subtraction are only for polynomials having
  the same :py:attr:`~ajdmom.poly.Poly.keyfor` attribute.

* The original poly or polys remain unchanged after these operations,
  all of which return a new poly.

Besides, I defined following methods:

+------------------------------------------+------------------------------------------------------------------+
|Method                                    |Function                                                          |
+==========================================+==================================================================+
|:py:meth:`~ajdmom.poly.Poly.merge`        |Merge with another poly having the same keyfor attribute          |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.add_keyval`   |Insert a new key-val or add val to the existing key               |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.remove_zero`  |Remove any item having value 0                                    |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.set_keyfor`   |Set the ``keyfor`` attribute for the poly                         |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.mul_poly`     |Multiply by another poly of different type and return a new one   |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.is_exact_type`|Check whether the supplied argument is a Poly with the same keyfor|
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.write_to_txt` |Write the poly to a txt file with the ``keyfor`` as the title line|
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.write_to_csv` |Write the poly to a csv file with the ``keyfor`` as the title line|
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.const_one`    |Constant equivalent to 1 in real numbers                          |
+------------------------------------------+------------------------------------------------------------------+
|:py:meth:`~ajdmom.poly.Poly.const_zero`   |Constant equivalent to 0 in real numbers                          |
+------------------------------------------+------------------------------------------------------------------+

Note that :py:meth:`~ajdmom.poly.Poly.merge` (in-place) is different from
the magic method :code:`__add__()` (not-in-place). The former changes the
original poly, while the latter does not.
"""
from collections import UserDict
from fractions import Fraction


def kv(i, key):
    """get key[i] where key is a tuple

    :param int i: index of key or -1 for not having the element.
    :param tuple key: tuple as a key for Poly.
    :return: key[i] if i in range(len(k)), otherwise 0.
    :rtype: int
    """
    if i in range(len(key)):
        return key[i]
    else:
        return 0


class Poly(UserDict):
    """Class for different versions of "Polynomial" """

    keyfor = ()
    """Indicating the purpose for each key component, a tuple of str."""

    def __add__(self, other):
        """Add another poly having the same keyfor."""
        if not self.is_exact_type(other):
            msg = "The operands must be Polys having the same 'keyfor' attribute."
            raise NotImplementedError(msg)
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly(self)  # UserDict initialization
        poly.set_keyfor(self.keyfor)
        #
        for k in other:
            poly.add_keyval(k, other[k])
        #
        poly.remove_zero()  # remove any item with 0 value
        return poly

    def merge(self, other):
        """Merge another poly having the same keyfor attribute.

        Insert new key-value or add value for the existing key.

        :param Poly other: poly having the same keyfor.
        :return: the updated first poly.
        :rtype: Poly
        """
        if not self.is_exact_type(other):
            msg = "The operands must be Polys having the same 'keyfor' attribute."
            raise NotImplementedError(msg)
        #
        for k in other:
            self.add_keyval(k, other[k])
        #
        self.remove_zero()  # remove any item with 0 value

    def __sub__(self, other):
        """Subtract another poly having the same keyfor."""
        if not self.is_exact_type(other):
            msg = "The operands must be Polys having the same 'keyfor' attribute."
            raise NotImplementedError(msg)
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly(self)  # UserDict initialization
        poly.set_keyfor(self.keyfor)
        #
        for k in other:
            if k in self:
                poly[k] -= other[k]
            else:
                poly[k] = -other[k]
        #
        poly.remove_zero()  # remove any item with 0 value
        return poly

    def __mul__(self, other):
        """Multiply by another poly."""
        if not self.is_exact_type(other):
            # msg = "The operands must be Polys having the same 'keyfor' attribute."
            # raise NotImplementedError(msg)
            kf1 = self.keyfor
            kf2 = other.keyfor
            kf = list(kf1)
            idx = []
            for k in kf2:
                if k not in kf1:
                    kf.append(k)
                    idx.append(len(kf) - 1)
                else:
                    idx.append(kf1.index(k))
            poly = Poly()
            poly.set_keyfor(kf)
            for k1, v1 in self.items():
                for k2, v2 in other.items():
                    key = list(k1) + [0 for i in range(len(kf) - len(kf1))]
                    for i in range(len(k2)):
                        key[idx[i]] += k2[i]
                    val = v1 * v2
                    poly.add_keyval(tuple(key), val)
            return poly
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly()
        poly.set_keyfor(self.keyfor)
        #
        for k1 in self:
            for k2 in other:
                key = tuple(k1[i] + k2[i] for i in range(len(k1)))
                val = self[k1] * other[k2]
                poly.add_keyval(key, val)
        #
        poly.remove_zero()  # remove any item with 0 value
        return poly

    def __rmul__(self, c):
        """Reversely multiply by a constant."""
        # Should we consider the special case c = 0?
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly()
        poly.set_keyfor(self.keyfor)
        #
        for k in self:
            val = self[k] * c
            poly.add_keyval(k, val)
        #
        poly.remove_zero()  # remove any item with 0 value
        return poly

    def __pow__(self, n):
        """Raise the poly to power n."""
        if n < 0:
            raise NotImplementedError("n must be an integer and >= 0.")
        # support for poly^0 = 1
        if n == 0:
            return self.const_one()
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly(self)
        poly.set_keyfor(self.keyfor)
        #
        for i in range(n - 1):
            poly = poly * self
        return poly

    def add_keyval(self, key, val):
        """Insert a new key-val or add val to the existing key.

        :param key: key for the poly, a tuple or list (will be converted to tuple).
        :type key: tuple or list
        :param Fraction val: corresponding value.
        """
        if isinstance(key, list): key = tuple(key)
        if key in self:
            self[key] += val
        else:
            self[key] = val

    def remove_zero(self):
        """Remove any item having value 0."""
        # val is n/m, fraction number
        ks = [k for k in self if self[k] == 0]  # works for fraction number
        for k in ks:
            del self[k]

    def set_keyfor(self, names):
        """Set the ``keyfor`` attribute for the poly.

        :param list names: a sequence of names,
           each corresponding to the key tuple counterpart,
           i.e., names[i] v.s. key[i].
        :type names: list or tuple
        """
        self.keyfor = tuple(names)

    def mul_poly(self, other, keyIndexes, keyfor):
        """Multiply by another poly of different type and return a new poly.

        :param Poly other: another poly having a different 'keyfor' attribute.
        :param tuple keyIndexes: a tuple with two lists,
           keyIndexes[0] for self,
           keyIndexes[1] for other.
        :param list keyfor: the 'keyfor' attribute for the returned poly.
        :return: a poly with 'keyfor' attribute.
        :rtype: Poly
        """
        #
        # initialize a new poly to isolate the operation from affecting
        # the original one
        #
        poly = Poly()
        poly.set_keyfor(keyfor)
        #
        idx1, idx2 = keyIndexes
        kN = len(idx1)
        #
        for k1 in self:
            for k2 in other:
                key = tuple(kv(idx1[i], k1) + kv(idx2[i], k2) for i in range(kN))
                val = self[k1] * other[k2]
                poly.add_keyval(key, val)
        #
        poly.remove_zero()
        return poly

    def is_exact_type(self, other):
        """Check whether the supplied argument is a poly having the same keyfor."""
        flag = True
        if not isinstance(other, Poly):
            flag = False
        else:
            try:
                if len(self.keyfor) != len(other.keyfor):
                    flag = False
                else:
                    for i in range(len(self.keyfor)):
                        if self.keyfor[i] != other.keyfor[i]:
                            flag = False
                            break
            except AttributeError:
                flag = False
        return flag

    def write_to_txt(self, filename):
        """Write the poly to a txt file with the ``keyfor`` as the title line."""
        # applies to all models: mdl_1fsv,...,mdl_svcj
        if not filename.endswith(".txt"): filename += ".txt"
        #
        title = f"{str(self.keyfor)}: value(in fraction)\n"
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(title)
            for k, v in self.items():
                f.write(f"{str(k)}: ({str(v.numerator)},{str(v.denominator)})\n")
        print(f"complete the writing of {filename}.")

    def write_to_csv(self, filename):
        """Write the poly to a csv file with the ``keyfor`` as the title line."""
        if not filename.endswith(".csv"): filename += ".csv"
        #
        # for cases without 'l_{1:n}',...: mdl_1fsv, mdl_1fsvj, mdl_2fsv, mdl_2fsvj
        #
        if 'l_{1:n}' not in self.keyfor:
            title = ','.join(self.keyfor) + ',num,den\n'
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(title)
                for k, v in self.items():
                    num, den = v.numerator, v.denominator
                    row = ','.join([str(s) for s in k]) + f',{str(num)},{str(den)}\n'
                    f.write(row)
            print(f"complete the writing of {filename}.")
            return
        #
        # find the index for 'l_{1:n}' in keyfor
        #
        i = 0
        while i < len(self.keyfor):
            if self.keyfor[i] == 'l_{1:n}':
                break
            else:
                i += 1
        # find one sub-key corresponding to 'l_{1:n}'
        first_key = next(iter(self))
        l = first_key[i]
        #
        # It is assumed that all sub-keys corresponding to 'l_{1:n}' have the same length
        #
        # make sure the assumption is already meet
        for k in self.keys():
            if len(k[i]) != len(l):
                msg = "The sub-keys corresponding to 'l_{1:n}' have different lengths!"
                msg += "You should first classify the poly into sub-polys according to the lengths."
                msg += "Then write each of the sub-polys to csv files."
                raise Exception(msg)
        #
        # for cases 'l_{1:m}', 'o_{2:n}': mdl_srjd
        #
        if 'p_{2:n}' not in self.keyfor:
            sub1 = ['l' + str(j) for j in range(1, len(l) + 1)]
            sub2 = ['o' + str(j) for j in range(2, len(l) + 1)]
            title = ','.join(self.keyfor[:i]) + ',' + ','.join(sub1 + sub2) + ','
            title += ','.join(self.keyfor[i + 2:]) + ',num,den\n'
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(title)
                for k, v in self.items():
                    row = ','.join([str(s) for s in k[:i] + k[i] + k[i + 1] + k[i + 2:]])
                    row += f',{v.numerator},{v.denominator}\n'
                    f.write(row)
        #
        # for cases 'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}': mdl_svvj, mdl_svij, mdl_svcj
        #
        else:
            sub1 = ['l' + str(j) for j in range(1, len(l) + 1)]
            sub2 = ['o' + str(j) for j in range(1, len(l) + 1)]
            sub3 = ['p' + str(j) for j in range(2, len(l) + 1)]
            sub4 = ['q' + str(j) for j in range(2, len(l) + 1)]
            title = ','.join(self.keyfor[:i]) + ',' + ','.join(sub1 + sub2 + sub3 + sub4) + ','
            title += ','.join(self.keyfor[i + 4:]) + ',num,den\n'
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(title)
                for k, v in self.items():
                    row = ','.join([str(s) for s in k[:i] + k[i] + k[i + 1] + k[i + 2] + k[i + 3] + k[i + 4:]])
                    row += f',{v.numerator},{v.denominator}\n'
                    f.write(row)
        print(f"complete the writing of {filename}.")

    def const_one(self):
        """Constant equivalent to 1 in real numbers.

        :return: poly evaluate to 1.
        :rtype: Poly
        """
        poly = Poly()
        poly.set_keyfor(self.keyfor)
        if isinstance(self.keyfor, tuple):
            key = tuple(0 for i in range(len(self.keyfor)))
        else:  # n must = 1 and keyfor is not a tuple
            key = 0
        val = Fraction(1, 1)
        poly.add_keyval(key, val)
        return poly

    def const_zero(self):
        """Constant equivalent to 0 in real numbers.

        :return: poly evaluate to 0.
        :rtype: Poly
        """
        poly = Poly()
        poly.set_keyfor(self.keyfor)
        if isinstance(self.keyfor, tuple):
            key = tuple(0 for i in range(len(self.keyfor)))
        else: # n must = 1 and keyfor is not a tuple
            key = 0
        val = Fraction(0, 1)
        poly.add_keyval(key, val)
        return poly


if __name__ == "__main__":
    # Example usage of class Poly, see 'tests/test_poly.py' for more test
    print('\nExample usage of class Poly\n')
    #
    pol1 = Poly({(1, 0): 1, (0, 1): 1})
    pol1.set_keyfor(['e', 'h'])
    pol2 = Poly({(1, 0): 1, (0, 2): 1})
    pol2.set_keyfor(['e', 'h'])
    print(f'Before addition,\npol1 = {pol1}\npol2 = {pol2}\n')
    print(f'pol1 + pol2 = {pol1 + pol2}\n')
    print(f'After addition,\npol1 = {pol1}\npol2 = {pol2}')

    pol3 = Poly({(1, 0, 1): 1})
    pol3.set_keyfor(['e', 'k', 'theta'])
    poln = pol1 * pol3
    print(f'\npol3 = {pol3}')
    print(f'poln = pol1 * pol3, \npoln = {poln}')
    print(f'poln.keyfor = {poln.keyfor}')



"""
Module for generating samples from :abbr:`SRJD(Square-Root Jump Diffusion)`

Note that we are generating i.i.d. samples with fixed conditions instead of
a trajectory of samples!
"""
import math
import numpy as np

rng = np.random.default_rng()


def rSRD(v0, tau, k, theta, sigma):
    r"""Generate next Random sample from Square-Root Diffusion

    :param float v0: current state value, > 0
    :param float tau: time difference between current and next state
    :param float k: parameter :math:`k`
    :param float theta: parameter :math:`\theta`
    :param float sigma: parameter :math:`\sigma`
    :return: next sample
    :rtype: float
    """
    d = 4 * theta * k / (sigma ** 2)  # degrees of freedom
    c = (sigma ** 2) * (1 - math.exp(-k * tau)) / (4 * k)  # leading constant
    lmbd = v0 * math.exp(-k * tau) / c  # non-centrality parameter
    return c * rng.noncentral_chisquare(d, lmbd)


def rSRJD(v0, k, theta, sigma, T, jumpsize, jumptime, n):
    r"""Generate i.i.d. Random Samples from mdl_srjd given v0 and jumps

    :param float v0: value of the initial state :math: `v_0 > 0`
    :param float k: parameter :math:`k`
    :param float theta: parameter :math:`\theta`
    :param float sigma: parameter :math:`\sigma`
    :param float T: end time
    :param tuple jumpsize: jump sizes over the interval [0,h]
    :param tuple jumptime: jump time points over the interval [0,h]
    :param int n: number of samples to generate
    :return: i.i.d. samples
    :rtype: numpy.ndarray
    """
    x = np.empty(n)
    if not isinstance(jumpsize, tuple):
        jumpsize = tuple(jumpsize)
    if not isinstance(jumptime, tuple):
        jumptime = tuple(jumptime)
    numJ = len(jumpsize)
    if numJ == 0:  # diffusion only
        for i in range(n): x[i] = rSRD(v0, T, k, theta, sigma)
        return x
    #
    # now numJ  >= 1
    #
    dt = (jumptime[0],) + tuple(jumptime[i] - jumptime[i - 1] for i in range(1, numJ))
    v0_original = v0
    for i in range(n):
        v0 = v0_original
        for j in range(numJ):
            vt = rSRD(v0, dt[j], k, theta, sigma) + jumpsize[j]
            v0 = vt
        # residual time
        t_residual = T - jumptime[-1]
        if t_residual > 0: vt = rSRD(v0, t_residual, k, theta, sigma)
        x[i] = vt
    return x


if __name__ == "__main__":
    import sys

    v0, h = 0.007569, 1
    k, theta, sigma = 3.46, 0.008, 0.14
    mu_v, lmbd = 0.05, 0.47
    #
    args = sys.argv[1:]
    numJ = rng.poisson(lmbd * h) if len(args) == 0 else int(args[0])
    jumpsize = rng.exponential(mu_v, numJ)
    jumptime = rng.uniform(0, h, numJ)
    jumptime = sorted(jumptime)
    #
    n = 20
    x = rSRJD(v0, k, theta, sigma, h, jumpsize, jumptime, n)
    print(f"Given v0 = {v0} and {numJ} jumps, ")
    print(f"jumpsize = {tuple(jumpsize)}, \njumptime = {tuple(jumptime)}")
    print(f"\n{n} replicate samples from SRJD: ")
    print(x)



"""
Utilities for combination and normal distribution
"""
import math
from fractions import Fraction as Frac
from ajdmom.poly import Poly


# ----------------------------combination----------------------------------------
def comb(n, ns):
    r"""number of occurrences for an exact combination

    :param int n: total power, = ``sum(ns)``
    :param list ns: list of powers, :math:`n_1 + \cdots + n_N = n`
    :return: number of occurrences for this exact combination
    :return: float
    """
    f = 1
    for k in ns:
        f *= math.comb(n, k)
        n -= k
    return f


def multilevel_index(n, m, show=False):
    """get the indexes of multiple summation

    :param int n: levels of the multiple summation, >= 1
    :param int m: number of possibilities in each summation, >= 1
    :param bool show: whether show the process or not, default to ``False``
    :return: list of the multi-level indexes.
    :rtype: list
    """
    index = []
    stack = []
    nextstep = "forward"  # downward|backward
    while len(stack) < n:
        if nextstep == "forward":
            if len(stack) < n - 1:
                stack.append(0)
                nextstep = "forward"
            else:
                for i in range(m):  # adapt to python 0-indexing
                    stack.append(i)
                    index.append(stack.copy())
                    if show: print(stack)
                    stack.pop()
                # checking whether all indexes traversed
                all_m = True
                for i in range(n - 1):  # no need to test the innermost level
                    if stack[i] != m - 1:
                        all_m = False
                        break
                if all_m: break
                nextstep = "backward"
        else:
            i = stack.pop()
            if i == m - 1:
                nextstep = "backward"
            else:
                stack.append(i + 1)
                nextstep = "forward"
    return index


def multilevel_comb(n, m, show=False):
    """get the indexes of multiple level combination

    :param int n: parts to split the number
    :param int m: total number resource
    :param bool show: whether show the process or not, default to ``False``
    :return: list of the multiple level combination indexes
    :return: list
    """
    if n == 1:
        return m
    elif n == 2:
        index = [[i, m - i] for i in range(m+1)]
        if show: print(index)
        return index
    # N > 2
    index = []
    stack = []
    nextstep = "forward"
    while len(stack) < n:
        if nextstep == "forward":
            if len(stack) < n - 2:
                stack.append(0)
                nextstep = "forward"
            else:
                left = m - sum(stack)
                for i in range(left + 1):
                    stack.append(i)
                    stack.append(left - i)
                    index.append(stack.copy())
                    if show: print(stack)
                    stack.pop()
                    stack.pop()
                # checking whether all indexes traversed
                if stack[0] == m: break
                #
                nextstep = "backward"
        else:
            left = m - sum(stack)
            if left == 0:
                stack.pop()
                nextstep = "backward"
            else:
                stack[-1] += 1
                nextstep = "forward"
    return index


# -----------------------------norm distribution---------------------------------
def dbfactorial(n):
    r"""double factorial :math:`n!!`

    :param int n: number :math:`n`
    :return: value of :math:`n!! = n(n-2)(n-4)\cdots 4 \cdot 2` for
      even :math:`n`, :math:`n!! = n(n-2)(n-4)\cdots 3 \cdot 1` for odd :math:`n`
    :rtype: float
    """
    if n == 0 or n == 1:
        return (1)
    elif n == 2:
        return (2)
    #
    f = n
    while n > 2:
        n -= 2
        f *= n
    return f


def cmnorm(n):
    """central moment of Normal distribution

    :param int n: order of the central moment, n >= 0
    :return: poly with attribute ``keyfor`` = ('sigma'), noting that its key
      is still a tuple, though it contains only one element
    :rtype: Poly
    """
    # special case
    if n == 0:
        P1 = Poly({(0,): Frac(1, 1)})
        P1.set_keyfor(['sigma'])
        return P1
    # n >= 1
    poly = Poly()
    poly.set_keyfor(['sigma'])
    if n % 2 != 0:
        poly.add_keyval((0,), Frac(0, 1))
    else:
        poly.add_keyval((n,), Frac(dbfactorial(n - 1), 1))
    return poly


# --------------------------------misc-------------------------------------------
# def fZ_old_each_index(l, o, p, q, k, s, J, index):
#     r"""function value :math:`f_{Z_t}(\boldsymbol{l},\boldsymbol{o},
#     \boldsymbol{p}, \boldsymbol{q})` as defined in :eq:`fZ` for an index
#
#     :param tuple l: vector :math:`\boldsymbol{l}`, should be a tuple
#     :param tuple o: vector :math:`\boldsymbol{o}`, should be a tuple
#     :param tuple p: vector :math:`\boldsymbol{p}`, should be a tuple
#     :param tuple q: vector :math:`\boldsymbol{q}`, should be a tuple
#     :param float k: parameter :math:`k`
#     :param tuple s: jump time points
#     :param tuple J: jump sizes
#     :param list index: an exact index combination
#     :return: function value for an exact index combination
#     :rtype: float
#     """
#     f = 1
#     s_max = s[index[0]]
#     for i in range(len(index)):
#         j = index[i]
#         s_j = s[j]
#         f *= math.exp(l[i] * k * s_j) * J[j] * (s_j ** o[i])
#         if i > 0:
#             s_max = s_max if s_max > s_j else s_j
#             f *= math.exp(p[i - 1] * k * s_max) * (s_max ** q[i - 1])
#     return f
#
#
# def fZ_old(l, o, p, q, k, s, J):
#     r"""function :math:`f_{Z_t}(\boldsymbol{l},\boldsymbol{o},\boldsymbol{p},
#     \boldsymbol{q})` as defined in :eq:`fZ`
#
#     :param tuple l: vector :math:`\boldsymbol{l}`, should be a tuple
#     :param tuple o: vector :math:`\boldsymbol{o}`, should be a tuple
#     :param tuple p: vector :math:`\boldsymbol{p}`, should be a tuple
#     :param tuple q: vector :math:`\boldsymbol{q}`, should be a tuple
#     :param float k: parameter :math:`k`
#     :param tuple s: jump time points
#     :param tuple J: jump sizes
#     :return: function value
#     :rtype: float
#     """
#     # 1. Continuous part
#     if len(l) == 0: return 1
#     #
#     # 2. Jump part, len(l) >= 1, may be single|multiple summation
#     # 2.1 no jump occurs
#     if len(J) == 0: return 0
#     # 2.2 >=1 jumps occur
#     val = 0
#     stack = []
#     nextstep = "forward"  # downward|backward
#     while len(stack) < len(l):
#         if nextstep == "forward":
#             if len(stack) < len(l) - 1:
#                 stack.append(0)
#                 nextstep = "forward"
#             else:
#                 # come to the innermost-level summation
#                 for i in range(len(J)):  # index start with 0
#                     stack.append(i)
#                     index = stack.copy()
#                     stack.pop()
#                     # evaluate and add to sum
#                     f = 1
#                     s_max = s[index[0]]
#                     for ii in range(len(index)):
#                         j = index[ii]
#                         s_j = s[j]
#                         f *= math.exp(l[ii] * k * s_j) * J[j] * (s_j ** o[ii])
#                         if ii > 0:
#                             s_max = s_max if s_max > s_j else s_j
#                             f *= math.exp(p[ii - 1] * k * s_max) * (s_max ** q[ii - 1])
#                     val += f
#                     #
#                 # checking whether all indexes traversed
#                 all_N = True
#                 for i in range(len(l) - 1):
#                     if stack[i] != len(J) - 1:
#                         all_N = False
#                         break
#                 if all_N: break
#                 #
#                 nextstep = "backward"
#         else:  # nextstep = "backward"
#             i = stack.pop()
#             if i == len(J) - 1:
#                 nextstep = "backward"
#             else:
#                 stack.append(i + 1)
#                 nextstep = "forward"
#     return val


def fZ(l, o, k, s, J):
    r"""function :math:`f_{Z_t}(\boldsymbol{l},\boldsymbol{o})` as defined in :eq:`fZ`

    :param tuple l: vector :math:`\boldsymbol{l}`, should be a tuple
    :param tuple o: vector :math:`\boldsymbol{o}`, should be a tuple
    :param float k: parameter :math:`k`
    :param tuple s: jump time points
    :param tuple J: jump sizes
    :return: function value
    :rtype: float
    """
    # 1. Continuous part
    if len(l) == 0: return 1
    #
    # 2. Jump part, len(l) >= 1, may be single|multiple summation
    # 2.1 no jump occurs
    if len(J) == 0: return 0
    # 2.2 >=1 jumps occur
    val = 0
    stack = []
    nextstep = "forward"  # downward|backward
    while len(stack) < len(l):
        if nextstep == "forward":
            if len(stack) < len(l) - 1:
                stack.append(0)
                nextstep = "forward"
            else:
                # come to the innermost-level summation
                for i in range(len(J)):  # index start with 0
                    stack.append(i)
                    index = stack.copy()
                    stack.pop()
                    # evaluate and add to sum
                    f = 1
                    s_max = s[index[0]]
                    for ii in range(len(index)):
                        j = index[ii]
                        s_j = s[j]
                        s_max = s_max if s_max > s_j else s_j
                        f *= math.exp(k * s_j) * J[j]
                        f *= math.exp(l[ii] * k * s_max) * (s_max ** o[ii])
                    val += f
                # checking whether all indexes traversed
                all_N = True
                for i in range(len(l) - 1):
                    if stack[i] != len(J) - 1:
                        all_N = False
                        break
                if all_N: break
                #
                nextstep = "backward"
        else:  # nextstep = "backward"
            i = stack.pop()
            if i == len(J) - 1:
                nextstep = "backward"
            else:
                stack.append(i + 1)
                nextstep = "forward"
    return val


def simplify_rho(poly, n):
    """expand the term 'sqrt(1-rho^2)'

    :param Poly poly: the poly contains 'sqrt(1-rho^2)' which must follow 'rho' immediately
    :param int n: index of 'sqrt(1-rho^2)'
    :return: poly without 'sqrt(1-rho^2)'
    :rtype: Poly
    """
    # exclude the 'sqrt(1-rho^2)' in the keyfor attribute
    kf = poly.keyfor[0:n] + poly.keyfor[(n + 1):]
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k in poly:
        if k[n] == 0:
            key = k[0:n] + k[(n + 1):]
            poln.add_keyval(key, poly[k])
        elif k[n] % 2 == 0:
            # (1-rho^2)^p = \sum_{i=0}^p C_p^i * (-1)^i * 1^{p-i} * rho^{2*i}
            p = k[n] // 2
            for i in range(0, p + 1):
                key = k[0:n] + k[(n + 1):]
                key = list(key)
                key[n - 1] += 2 * i
                key = tuple(key)
                val = poly[k] * math.comb(p, i) * ((-1) ** i)
                poln.add_keyval(key, val)
        else:
            raise Exception("can not reduce the 'sqrt(1-rho^2)' dimension!")
    poln.remove_zero()
    return poln


if __name__ == "__main__":
    from pprint import pprint

    print(f"comb(5, [1, 2, 2) = {comb(5, [1, 2, 2])}")
    print("\nThe index combinations of 3 levels of summation, each 2 possibilities: ")
    multilevel_index(3, 2, show=True)
    print("\nThe combinations of split 4 identical items into 3 parts: ")
    multilevel_comb(3, 4, show=True)
    print("\ncmnorm() returns a poly with keyfor = ('sigma'): ")
    for i in range(7):
        print(f"cmnorm({i}) = ")
        pprint(cmnorm(i))



"""
Module for generating a trajectory of samples from mdl_1fsvj by
Euler approximation

To facilitate the verification of the correctness of our codes
by comparison the population moments (given by our package) and
their sample counterparts.
Here I define function :py:func:`~ajdmom.mdl_1fsvj.euler.r1FSVJ`.
"""
import math
import numpy as np


def r1FSVJ(v0, par, N, n_segment=10, h=1):
    """Generate samples from mdl_1fsvj by Euler approximation

    :param float v0: value of the initial variance :math:`v(0) > 0`.
    :param dict par: parameters in a dict.
    :param int N: target length of samples to generate.
    :param int n_segment: number of segments each interval is split into.
    :param float h: time interval between any two consecutive samples.
    :return: a sequence of samples.
    :rtype: numpy 1-D array.
    """
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma_v']
    rho = par['rho']
    lmbd = par['lambda']
    mu_j = par['mu_j']
    sigma_j = par['sigma_j']
    #
    dlt = h / n_segment
    std = math.sqrt(dlt)
    #
    y_series = np.empty(N)
    #
    for i in range(N):
        I = 0
        I2 = 0
        IV = 0
        for j in range(n_segment):
            eps = math.sqrt(v0) * np.random.normal(0, std)
            eps2 = math.sqrt(v0) * np.random.normal(0, std)
            v = v0 + k * (theta - v0) * dlt + sigma * eps
            if v < 0:
                v = 0  # eps = -(v0 + k*(theta-v0)*dlt)/sigma
            I += eps
            I2 += eps2
            IV += v0 * dlt
            v0 = v
        #
        # jump part
        n = np.random.poisson(lmbd * h)
        J_n = 0
        if n > 0:
            for l in range(n):
                J_n += np.random.normal(mu_j, sigma_j)
        #
        y = mu * h - IV / 2 + rho * I + math.sqrt(1 - rho ** 2) * I2 + J_n
        y_series[i] = y
    #
    return y_series



"""
Conditional Moments of the SVCJ model, given initial variance
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.utils import comb
from ajdmom.mdl_svcj.cond_ieii_ieziziz_mom import moment_ieii_ieziziz


def b_n(n1, n2, n3, n4, n5, n6, n7, n8):
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho', 'sqrt(1-rho^2)',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    poly = Poly()
    poly.set_keyfor(kf)
    for i1 in range(n2 + 1):
        for i2 in range(n5 + 1):
            for i3 in range(n7 + 1):
                for i4 in range(n8 + 1):
                    bino1 = math.comb(n2, i1)
                    bino2 = math.comb(n5, i2)
                    bino3 = math.comb(n7, i3)
                    bino4 = math.comb(n8, i4)
                    bino = bino1 * bino2 * bino3 * bino4
                    sign = (-1) ** (i1 + i2 + i3 + i4)
                    c = sign * Fraction(bino, 2 ** (n1 + n4 + i1 + i2 + i3 + n8))
                    #
                    key = (-(n1 + n4 + n8 - i4), n7, n1 + n4 + i1 + i2 + n8,
                           n7 - i3, n8, i3, n1 + i1, n2 - i1, n3,
                           0, 0, n5 - i2, 0, 0)
                    poly.add_keyval(key, c)
    return poly


def enlarge(poly, new_keyfor):
    # it requires n < m
    n, m = len(poly.keyfor), len(new_keyfor)
    index = []  # indexes of elements of poly.keyfor in new_keyfor
    for k in poly.keyfor:
        index.append(new_keyfor.index(k))
        # if not found, it will raise an error
    poln = Poly()
    poln.set_keyfor(new_keyfor)
    for k, v in poly.items():
        key = [0 for i in range(m)]
        for i in range(n):
            key[index[i]] = k[i]
        poln.add_keyval(tuple(key), v)
    return poln


def expand_sqrt(poly):
    """expand sqrt(1-rho^2) in poly key"""
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'v0-theta', 'theta', 'sigma', 'rho', 'sqrt(1-rho^2)',
    #       'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        if k[8] % 2 != 0:
            raise Exception("the power of sqrt(1-rho^2) is not even!")
        for i in range(k[8] // 2 + 1):
            key = list(k)
            key[7] += 2 * i
            del key[8]
            key = tuple(key)
            sign = (-1) ** i
            val = sign * math.comb(k[8] // 2, i) * v
            poln.add_keyval(key, val)
    return poln


def moment_y(n):
    """conditional moment of :math:`y_t|v_0`

    :param integer n: order of the conditional moment
    :return: poly with ``keyfor`` = ('e^{kt}', 't', 'k^{-}',
      'mu', 'v0-theta', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s')
    :rtype: Poly
    """
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    if n < 0:
        raise ValueError('n must be non-negative')
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    # special case
    if n == 0:
        return p1
    # typical cases
    kf.insert(8, 'sqrt(1-rho^2)')
    poly = Poly()
    poly.set_keyfor(kf)
    for n1 in range(n + 1):
        for n2 in range(n - n1 + 1):
            for n3 in range(n - n1 - n2 + 1):
                for n4 in range(n - n1 - n2 - n3 + 1):
                    for n5 in range(n - n1 - n2 - n3 - n4 + 1):
                        for n6 in range(n - n1 - n2 - n3 - n4 - n5 + 1):
                            for n7 in range(n - n1 - n2 - n3 - n4 - n5 - n6 + 1):
                                n8 = n - n1 - n2 - n3 - n4 - n5 - n6 - n7
                                c = comb(n, [n1, n2, n3, n4, n5, n6, n7, n8])
                                # print(f"({n1}, {n2}, {n3}, {n4}, {n5}, {n6}, {n7}, {n8}), c = {c}")
                                pol1 = b_n(n1, n2, n3, n4, n5, n6, n7, n8)
                                # kf = ['e^{kt}', 't', 'k^{-}',
                                #       'mu', 'v0-theta', 'theta', 'sigma', 'rho', 'sqrt(1-rho^2)',
                                #        'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
                                # print("b_n = "); pprint(pol1); print(f"{pol1.keyfor}")
                                pol2 = moment_ieii_ieziziz(n1, n2, n3, n4, n5, n6)
                                # kf = ['e^{kt}','t','k^{-}',
                                #       'v0-theta','theta','sigma',
                                #       'lmbd','mu_v','mu_s','sigma_s']
                                # print("moment_ieii_ieziziz = "); pprint(pol2); print(f"{pol2.keyfor}\n")
                                pol2 = enlarge(pol2, pol1.keyfor)
                                poly.merge(c * (pol1 * pol2))
    return expand_sqrt(poly)  # expand sqrt(1-rho^2)


def poly2num(poly, par):
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    k = par['k']
    mu, v0, theta, sigma, rho = par['mu'], par['v0'], par['theta'], par['sigma'], par['rho']
    lmbd, mu_v, rhoJ, mu_s, sigma_s = par['lmbd'], par['mu_v'], par['rhoJ'], par['mu_s'], par['sigma_s']
    h = par['h']
    #
    f = 0
    for K, V in poly.items():
        val = math.exp(K[0] * k * h) * (h ** K[1]) / (k ** K[2])
        val *= (mu ** K[3]) * ((v0 - theta) ** K[4]) * (theta ** K[5])
        val *= (sigma ** K[6]) * (rho ** K[7])
        val *= (lmbd ** K[8]) * (mu_v ** K[9]) * (rhoJ ** K[10])
        val *= (mu_s ** K[11]) * (sigma_s ** K[12])
        f += val * V
    return f


def m_y(n, par):
    moment = moment_y(n)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    n = 3
    poly = moment_y(n)
    # print(f"moment_y({n}) = "); pprint(poly); print(f"which is a poly with keyfor = {poly.keyfor}")

    v0 = 0.007569
    k = 3.46
    theta = 0.008
    sigma = 0.14
    rho = -0.82
    r = 0.0319
    lmbd = 0.47
    mu_b = -0.1
    sigma_s = 0.0001
    mu_v = 0.05
    rhoJ = -0.38

    mu = r - lmbd * mu_b
    mu_s = math.log((1 + mu_b) * (1 - rhoJ * mu_v)) - sigma_s ** 2 / 2
    h = 1

    par = {'v0': v0, 'mu': mu, 'k': k, 'theta': theta, 'sigma': sigma, 'rho': rho,
           'lmbd': lmbd, 'mu_v': mu_v, 'rhoJ': rhoJ, 'mu_s': mu_s, 'sigma_s': sigma_s,
           'h': h}

    for n in range(1, 4+1):
        print(f"moment({n}) = {m_y(n, par)}")



r"""
Conditional Central Moment
----------------------------

Conditional central moments of the Heston SV model, given the initial variance.

Note that the keyfor attribute is different from that in the module
:py:mod:`ajdmom.mdl_1fsv.cmom`, which now is
('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'sigma/2k',
'rho-sigma/2k', 'sqrt(1-rho^2)')

.. math::

   y_{n-1,t} = (\mu - \theta/2)[t-(n-1)h] - \beta_{n-1,t}(v_{n-1}-\theta)
    + \bar{y}_{n-1,t}

where :math:`\beta_{n-1,t} = (1-e^{-k[t-(n-1)h]})/(2k)` and the centralized term

.. math::

   \bar{y}_{n-1,t}
   = \frac{\sigma_v}{2k}e^{-kt}I\!E_{n-1,t} +
   \left( \rho - \frac{\sigma_v}{2k} \right)I_{n-1,t} +
   \sqrt{1-\rho^2}I_{n-1,t}^{*}.

The module implements the derivation of conditional moments of the centralized return

.. math::

   \mathbb{E}[\bar{y}_{n-1,t}^m|v_{n-1}]
   = \sum_{m_1+m_2+m_3=m} c(\boldsymbol{m}) b(\boldsymbol{m})
   (e^{-kt}I\!E_{n-1,t})^{m_1}(I_{n-1,t})^{m_2}(I_{n-1,t}^{*})^{m_3}

with :math:`c(\boldsymbol{m}) = C_m^{m_1}C_{m-m_1}^{m_2}`,
:math:`b(\boldsymbol{m}) =
\left(\frac{\sigma_v}{2k}\right)^{m_1}
\left(\rho - \frac{\sigma_v}{2k}\right)^{m_2}
\left(\sqrt{1-\rho^2}\right)^{m_3}`
and
:math:`\boldsymbol{m} = (m_1,m_2,m_3)`.

"""
import math
from fractions import Fraction as Frac

from ajdmom.utils import comb, simplify_rho
from ajdmom.poly import Poly
from ajdmom.ito_mom import int_et


def int_mIEII(c, tp, m, poly):
    r"""Integral of :math:`c \times tp \times \int_{(n-1)h}^t e^{mks} poly ds`

    :param int c: coefficient to multiply with
    :param int tp: type of the multiplication,

       +----+--------------------------------------+
       | tp | multiply with                        |
       +====+======================================+
       | 1  | :math:`e^{k(n-1)h}(v_{n-1}-\theta)`  |
       +----+--------------------------------------+
       | 2  | :math:`\theta`                       |
       +----+--------------------------------------+
       | 3  | :math:`\sigma_v`                     |
       +----+--------------------------------------+

    :param int m: power of :math:`e^{ks}` in the integrand
    :param Poly poly: a Poly object, such as
       :math:`\mathbb{E}[I\!E_t^{m_1} I_t^{m_2} I_t^{*m_3}]`, with attribute
       ``keyfor`` = ('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma').
    :return: poly with the same ``keyfor`` as that in the input poly
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(poly.keyfor) # poly = IEII[(m1, m2, m3)]
    for k1 in poly:  # ('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma')
        poly_sub = int_et(m + k1[0], k1[1])
        for k2 in poly_sub:  # ('e^{kt}', 't', 'k^{-}')
            key = [k2[0], k2[1], k2[2] + k1[2], k1[3], k1[4], k1[5]]
            if tp == 1:
                key[3] += 1
            elif tp == 2:
                key[4] += 1
            else:
                key[5] += 1
            val = c * poly[k1] * poly_sub[k2]
            poln.add_keyval(tuple(key), val)
    return poln


def recursive_IEII(n3, n4, n5, IEII):
    r"""Recursive step in equation :eq:`ito-moment`

    :param int n3: :math:`n_3`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n4: :math:`n_4`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n5: :math:`n_5`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param dict IEII: a dict with key (n3,n4,n5) and value Poly object with attribute
       ``keyfor`` = ('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma')
    :return: updated IEII.
    :rtype: dict
    """
    kf = ['e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    if n3 >= 2 and n4 >= 0 and n5 >= 0:
        c = Frac(n3 * (n3 - 1), 2)
        poly.merge(int_mIEII(c, 1, 1, IEII[(n3 - 2, n4, n5)]))
        poly.merge(int_mIEII(c, 2, 2, IEII[(n3 - 2, n4, n5)]))
        poly.merge(int_mIEII(c, 3, 1, IEII[(n3 - 1, n4, n5)]))
    if n3 >= 0 and n4 >= 2 and n5 >= 0:
        c = Frac(n4 * (n4 - 1), 2)
        poly.merge(int_mIEII(c, 1, -1, IEII[(n3, n4 - 2, n5)]))
        poly.merge(int_mIEII(c, 2, 0, IEII[(n3, n4 - 2, n5)]))
        poly.merge(int_mIEII(c, 3, -1, IEII[(n3 + 1, n4 - 2, n5)]))
    if n3 >= 1 and n4 >= 1 and n5 >= 0:
        c = Frac(n3 * n4, 1)
        poly.merge(int_mIEII(c, 1, 0, IEII[(n3 - 1, n4 - 1, n5)]))
        poly.merge(int_mIEII(c, 2, 1, IEII[(n3 - 1, n4 - 1, n5)]))
        poly.merge(int_mIEII(c, 3, 0, IEII[(n3, n4 - 1, n5)]))
    if n3 >= 0 and n4 >= 0 and n5 >= 2:
        c = Frac(n5 * (n5 - 1), 2)
        poly.merge(int_mIEII(c, 1, -1, IEII[(n3, n4, n5 - 2)]))
        poly.merge(int_mIEII(c, 2, 0, IEII[(n3, n4, n5 - 2)]))
        poly.merge(int_mIEII(c, 3, -1, IEII[(n3 + 1, n4, n5 - 2)]))
    return poly


def simplify(poly, tp=1):
    """Simplify polynomials differently

    :param Poly poly:  Poly with ``keyfor`` =
      ('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma',
      'sigma/2k', 'rho-sigma/2k', 'sqrt(1-rho^2)')
    :param int tp: type of the simplification.
     If ``tp=1`` (default), change 'e^{kt}' to 'e^{-kt}'.
     If ``tp=2``, change to
     ('e^{-kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'rho').
     If ``tp=3``, change to
     ('e^{-kt}', 't', 'k^{-}', 'v_0', 'theta', 'sigma', 'rho').
    :return: simplified poly.
    :rtype: Poly
    """
    if tp not in [1, 2, 3]: raise Exception("tp must be 1, 2, or 3")
    #
    kf = ('e^{-kt}',) + poly.keyfor[1:]
    poln = Poly()
    poln.set_keyfor(kf)
    #
    if tp == 1:  # change 'e^{kt}' to 'e^{-kt}'
        for k, v in poly.items():
            poln.add_keyval((-k[0],) + k[1:], v)
        return poln
    #
    if tp >= 2:  # change 'sigma/2k', 'rho-sigma/2k', 'sqrt(1-rho^2)' to 'rho'
        kf = ['e^{-kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'rho',
        'sqrt(1-rho^2)']
        poln.set_keyfor(kf)
        for k, v in poly.items():
            for i in range(0, k[7] + 1):  # expand 'rho-sigma/2k'
                p = k[6] + i
                key = (-k[0], k[1], k[2] + p, k[3], k[4], k[5] + p, k[7] - i, k[8])
                val = math.comb(k[7], i) * ((-1) ** i) * Frac(1, 2 ** p) * v
                poln.add_keyval(key, val)
        poln.remove_zero()
        poln = simplify_rho(poln, 7)
    #
    if tp == 3:
        # kf = ['e^{-kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'rho']
        # to
        # kf = ['e^{-kt}', 't', 'k^{-}', 'v_0', 'theta', 'sigma', 'rho']
        kf = ['e^{-kt}', 't', 'k^{-}', 'v_0', 'theta', 'sigma', 'rho']
        poly = Poly()
        poly.set_keyfor(kf)
        #
        for k, v in poln.items():
            for i in range(0, k[3] + 1):
                key = (k[0], k[1], k[2], k[3] - i, k[4] + i, k[5], k[6])
                val = v * math.comb(k[3], i) * ((-1) ** i)
                poly.add_keyval(key, val)
        poln = poly
    return poln


def cmoments_y_to(l, show=False):
    """Conditional central moments from order 1 to :math:`l`.

    Derive the conditional central moments and write to csv files.

    :param int l: highest order of the conditional central moments to derive.
    :param bool show: show the in-process message or not, defaults to False
    :return: a list of the conditional central moments with ``keyfor`` =
     ('e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma',
     'sigma/2k', 'rho-sigma/2k', 'sqrt(1-rho^2)')
    :rtype: list
    """
    # IEII: a dict of moments of E[IE^{n3}I^{n4}I^{*n5}]
    IEII = {}
    kf = ['e^{kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma']
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    P0 = Poly({(0, 0, 0, 0, 0, 0): Frac(0, 1)})
    P0.set_keyfor(kf)
    # n3 + n4 + n5 = 0
    # support for special case
    IEII[(0, 0, 0)] = P1
    # n3 + n4 + n5 = 1
    IEII[(1, 0, 0)] = P0
    IEII[(0, 1, 0)] = P0
    IEII[(0, 0, 1)] = P0
    # n3 + n4 + n5 = 2
    poly = Poly({(1, 0, 1, 1, 0, 0): Frac(1, 1),
                 (0, 0, 1, 1, 0, 0): -Frac(1, 1),
                 (2, 0, 1, 0, 1, 0): Frac(1, 2),
                 (0, 0, 1, 0, 1, 0): -Frac(1, 2)})
    poly.set_keyfor(kf)
    IEII[(2, 0, 0)] = poly
    poly = Poly({(0, 1, 0, 1, 0, 0): Frac(1, 1),
                 (1, 0, 1, 0, 1, 0): Frac(1, 1),
                 (0, 0, 1, 0, 1, 0): -Frac(1, 1)})
    poly.set_keyfor(kf)
    IEII[(1, 1, 0)] = poly
    IEII[(1, 0, 1)] = P0
    poly = Poly({(-1, 0, 1, 1, 0, 0): -Frac(1, 1),
                 (0, 0, 1, 1, 0, 0): Frac(1, 1),
                 (0, 1, 0, 0, 1, 0): Frac(1, 1)})
    poly.set_keyfor(kf)
    IEII[(0, 2, 0)] = poly
    IEII[(0, 1, 1)] = P0
    IEII[(0, 0, 2)] = poly
    #
    kf = kf + ['sigma/2k', 'rho-sigma/2k', 'sqrt(1-rho^2)']
    #
    cmoms = []
    #
    # special case: 0-th central moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    cmoms.append(P1)  # equiv to constant 1
    #
    # special case: 1-th central moment
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0): Frac(0, 1)})
    P0.set_keyfor(kf)
    cmoms.append(P0)  # equiv to constant 0
    #
    # typical cases
    #
    for n in range(2, l + 1):
        # container for the n-th conditional central moment of y
        poly = Poly()
        poly.set_keyfor(kf)
        # iter over all combinations of i + j + (n-i-j) = n
        for i in range(n, -1, -1):
            for j in range(n - i, -1, -1):
                # derive for this exact combination
                c = comb(n, [i, j, n - i - j])
                if n < 3:
                    poln = IEII[(i, j, n - i - j)]  # already in IEII
                else:
                    poln = recursive_IEII(i, j, n - i - j, IEII)
                # add this exact combination into poly
                # c * b * poln
                for k in poln:
                    # key = list(k)
                    # key[0] -= i
                    # key = tuple(key)
                    # key = key + (i, j, n - i - j)
                    key = (k[0]-i,) + k[1:] + (i, j, n - i - j)  # enlarge with last 3
                    poly.add_keyval(key, c * poln[k])
                # reserve poln for further use or delete it otherwise
                if n >= 3 and n < l:
                    IEII[(i, j, n - i - j)] = poln
                else:
                    del poln
        # delete polys when no more needed
        if n == l:
            del IEII
        else:  # only reserve the most recent polys
            index = [key for key in IEII if key[0] + key[1] + key[2] == n - 2]
            for key in index: del IEII[key]
        poly.remove_zero()
        cmoms.append(poly)
        if show: print(f"complete the derivation for the {n}-th conditional central moment.")
    return cmoms


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
     ('e^{-kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'rho').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    v0 = par['v0']
    k = par['k']
    h = par['h']
    theta = par['theta']
    sigma = par['sigma_v']
    rho = par['rho']
    #
    value = 0
    for K in poly:
        val = poly[K] * math.exp(-K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= ((v0 - theta) ** K[3]) * (theta ** K[4]) * (sigma ** K[5])
        val *= rho ** K[6]
        value += val
    return value

def cond_cm(l, par):
    """Conditional central moment in scalar

    :param int l: order of the conditional central moment.
    :param dict par: parameters in dict.
    :return: scalar of the conditional central moment.
    :rtype: float
    """
    cmoments = cmoments_y_to(l)
    cmoment = cmoments[-1]
    cmoment = simplify(cmoment, tp=2)
    # ('e^{-kt}', 't', 'k^{-}', 'v_0-theta', 'theta', 'sigma', 'rho')
    value = poly2num(cmoment, par)
    return value

if __name__ == "__main__":
    import sys
    from pprint import pprint

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l = 3 if len(args) == 0 else int(args[0])
    tp = 2 if len(args) <= 1 else int(args[1])
    cmoms = cmoments_y_to(l, show=True)  # 0-th to l-th conditional central moments
    kf = cmoms[0].keyfor
    print(f"\ncmoments_y_to(l) returns a list of polys with keyfor = \n{kf}")
    # for i in range(1, l+1):
    #   cmom = cmoms[i]
    #   cmom = simplify(cmom, tp)
    #   cmom.write_to_csv(f"tp{tp}-heston-cond-cmoment-{i}-formula.csv")
    print(f"\ncmoment_y({l}) = ")
    pprint(cmoms[l])
    print(f"which is a poly with keyfor = \n{kf}")
